import gymnasium as gym
import numpy as np

from metadrive.component.map.pg_map import PGMap
from metadrive.component.pgblock.bottleneck import Merge, Split
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.tollgate import TollGate
from metadrive.component.road_network import Road
from metadrive.constants import TerminationState
from metadrive.envs.marl_envs.multi_agent_metadrive import MultiAgentMetaDrive
from metadrive.manager.pg_map_manager import PGMapManager
from metadrive.obs.state_obs import LidarStateObservation, StateObservation
from metadrive.utils import Config, clip

MATollConfig = dict(
    spawn_roads=[Road(FirstPGBlock.NODE_2, FirstPGBlock.NODE_3), -Road(Merge.node(3, 0, 0), Merge.node(3, 0, 1))],
    num_agents=40,
    map_config=dict(exit_length=70, lane_num=3, toll_lane_num=8, toll_length=10),
    top_down_camera_initial_x=125,
    top_down_camera_initial_y=0,
    top_down_camera_initial_z=120,
    cross_yellow_line_done=True,
    # ===== Reward Scheme =====
    speed_reward=0.0,
    overspeed_penalty=0.5,
    vehicle_config=dict(
        min_pass_steps=30,  # We ask the agents to stop at tollgate for at least 6s (30 steps).
        show_lidar=False,
        # "show_side_detector": True,
        # "show_lane_line_detector": True,
        side_detector=dict(num_lasers=72, distance=20),  # laser num, distance
        lane_line_detector=dict(num_lasers=4, distance=20),
        lidar=dict(num_lasers=72, distance=20),
    )
)


class StayTimeManager:
    def __init__(self):
        self.entry_time = {}
        self.exit_time = {}
        self.last_block = {}

    def reset(self):
        self.entry_time = {}
        self.exit_time = {}
        self.last_block = {}

    def record(self, agents, time_step):
        for v_id, v in agents.items():
            cur_block_id = v.navigation.current_road.block_ID()
            if v_id in self.last_block:
                last_block_id = self.last_block[v_id]
                self.last_block[v_id] = cur_block_id
                if last_block_id != cur_block_id:
                    if cur_block_id == TollGate.ID:
                        # entry
                        self.entry_time[v_id] = time_step
                    elif (cur_block_id == Merge.ID or cur_block_id == Split.ID) and last_block_id == TollGate.ID:
                        self.exit_time[v_id] = time_step
            else:
                self.last_block[v_id] = cur_block_id


class TollGateStateObservation(StateObservation):
    # no intersection exclude navi info now
    @property
    def observation_space(self):
        # Navi info + Other states
        shape = self.ego_state_obs_dim + self.get_line_detector_dim()
        return gym.spaces.Box(-0.0, 1.0, shape=(shape, ), dtype=np.float32)

    def observe(self, vehicle):
        ego_state = self.vehicle_state(vehicle)
        return ego_state


class TollGateObservation(LidarStateObservation):
    def __init__(self, vehicle_config):
        super(LidarStateObservation, self).__init__(vehicle_config)
        self.state_obs = TollGateStateObservation(vehicle_config)
        self.in_toll_time = 0

    @property
    def observation_space(self):
        shape = list(self.state_obs.observation_space.shape)
        if self.config["vehicle_config"]["lidar"]["num_lasers"] > 0 \
                and self.config["vehicle_config"]["lidar"]["distance"] > 0:
            # Number of lidar rays and distance should be positive!
            shape[0] += self.config["vehicle_config"]["lidar"]["num_lasers"] \
                        + self.config["vehicle_config"]["lidar"]["num_others"] * 4 + 2
        return gym.spaces.Box(-0.0, 1.0, shape=tuple(shape), dtype=np.float32)

    def reset(self, env, vehicle=None):
        self.in_toll_time = 0

    def observe(self, vehicle):
        cur_block_is_toll = vehicle.navigation.current_road.block_ID() == TollGate.ID
        self.in_toll_time += 1 if cur_block_is_toll else 0
        if not cur_block_is_toll:
            toll_obs = [0.0, 0.0]
        else:
            toll_obs = [
                1.0 if cur_block_is_toll else 0.0, 1.0 if self.in_toll_time > vehicle.config["min_pass_steps"] else 0.0
            ]
        # print(toll_obs)
        state = self.state_observe(vehicle)
        other_v_info = self.lidar_observe(vehicle)
        ret = np.concatenate((state, np.asarray(other_v_info), np.asarray(toll_obs)))
        return ret.astype(np.float32)


class MATollGateMap(PGMap):
    BOTTLE_LENGTH = 35

    def _generate(self):
        length = self.config["exit_length"]

        parent_node_path, physics_world = self.engine.worldNP, self.engine.physics_world
        assert len(self.road_network.graph) == 0, "These Map is not empty, please create a new map to read config"

        # Build a first-block
        last_block = FirstPGBlock(
            self.road_network,
            self.config[self.LANE_WIDTH],
            self.config["lane_num"],
            parent_node_path,
            physics_world,
            length=length
        )
        self.blocks.append(last_block)

        split = Split(
            1, last_block.get_socket(index=0), self.road_network, random_seed=1, ignore_intersection_checking=False
        )
        split.construct_block(
            parent_node_path, physics_world, {
                "length": 2,
                "lane_num": self.config["toll_lane_num"] - self.config["lane_num"],
                "bottle_len": self.BOTTLE_LENGTH,
            }
        )
        self.blocks.append(split)
        toll = TollGate(
            2, split.get_socket(index=0), self.road_network, random_seed=1, ignore_intersection_checking=False
        )
        toll.construct_block(parent_node_path, physics_world, {
            "length": self.config["toll_length"],
        })

        self.blocks.append(toll)

        # Build Toll
        merge = Merge(3, toll.get_socket(index=0), self.road_network, random_seed=1, ignore_intersection_checking=False)
        merge.construct_from_config(
            dict(
                lane_num=self.config["toll_lane_num"] - self.config["lane_num"],
                length=self.config["exit_length"],
                bottle_len=self.BOTTLE_LENGTH,
            ), parent_node_path, physics_world
        )
        self.blocks.append(merge)


class MATollGatePGMapManager(PGMapManager):
    def reset(self):
        config = self.engine.global_config
        if len(self.spawned_objects) == 0:
            _map = self.spawn_object(MATollGateMap, map_config=config["map_config"], random_seed=None)
        else:
            assert len(self.spawned_objects) == 1, "It is supposed to contain one map in this manager"
            _map = self.spawned_objects.values()[0]
        self.load_map(_map)
        self.current_map.spawn_roads = config["spawn_roads"]


class MultiAgentTollgateEnv(MultiAgentMetaDrive):
    def __init__(self, config=None):
        super(MultiAgentTollgateEnv, self).__init__(config)
        self.stay_time_manager = StayTimeManager()

    def reset(self, *args, **kwargs):
        self.stay_time_manager.reset()
        return super(MultiAgentTollgateEnv, self).reset(*args, **kwargs)

    @staticmethod
    def default_config() -> Config:
        assert MATollConfig["vehicle_config"]["side_detector"]["num_lasers"] > 2
        assert MATollConfig["vehicle_config"]["lane_line_detector"]["num_lasers"] > 2
        return MultiAgentMetaDrive.default_config().update(MATollConfig, allow_add_new_key=True)

    def reward_function(self, vehicle_id: str):
        """
        Override this func to get a new reward function
        :param vehicle_id: id of BaseVehicle
        :return: reward
        """
        vehicle = self.vehicles[vehicle_id]
        step_info = dict()

        # Reward for moving forward in current lane
        if vehicle.lane in vehicle.navigation.current_ref_lanes:
            current_lane = vehicle.lane
        else:
            current_lane = vehicle.navigation.current_ref_lanes[0]
            current_road = vehicle.navigation.current_road
        long_last, _ = current_lane.local_coordinates(vehicle.last_position)
        long_now, lateral_now = current_lane.local_coordinates(vehicle.position)

        # reward for lane keeping, without it vehicle can learn to overtake but fail to keep in lane
        if self.config["use_lateral_reward"]:
            lateral_factor = clip(1 - 2 * abs(lateral_now) / vehicle.navigation.get_current_lane_width(), 0.0, 1.0)
        else:
            lateral_factor = 1.0

        reward = 0.0
        reward += self.config["driving_reward"] * (long_now - long_last) * lateral_factor

        if vehicle.navigation.current_road.block_ID() == TollGate.ID:
            if vehicle.overspeed:  # Too fast!
                reward = -self.config["overspeed_penalty"] * vehicle.speed_km_h / vehicle.max_speed_km_h
            else:
                # Good! At very low speed
                pass
        else:
            reward += self.config["speed_reward"] * (vehicle.speed_km_h / vehicle.max_speed_km_h)

        step_info["step_reward"] = reward

        if self._is_arrive_destination(vehicle):
            reward = +self.config["success_reward"]
        elif self._is_out_of_road(vehicle):
            reward = -self.config["out_of_road_penalty"]
        elif vehicle.crash_vehicle:
            reward = -self.config["crash_vehicle_penalty"]
        elif vehicle.crash_object:
            reward = -self.config["crash_object_penalty"]
        return reward, step_info

    def _is_out_of_road(self, vehicle):
        # A specified function to determine whether this vehicle should be done.
        # return vehicle.on_yellow_continuous_line or (not vehicle.on_lane) or vehicle.crash_sidewalk
        ret = vehicle.crash_sidewalk
        if self.config["cross_yellow_line_done"]:
            ret = ret or vehicle.on_yellow_continuous_line
        return ret

    def done_function(self, vehicle_id):
        done, done_info = super(MultiAgentMetaDrive, self).done_function(vehicle_id)
        if done_info[TerminationState.CRASH_VEHICLE] and (not self.config["crash_done"]):
            assert done_info[TerminationState.CRASH_VEHICLE] or done_info[TerminationState.CRASH_BUILDING] or \
                   done_info[TerminationState.SUCCESS] or done_info[TerminationState.OUT_OF_ROAD]
            if not (done_info[TerminationState.SUCCESS] or done_info[TerminationState.OUT_OF_ROAD]):
                # Does not revert done if high-priority termination happens!
                done = False

        if done_info[TerminationState.OUT_OF_ROAD] and (not self.config["out_of_road_done"]):
            assert done_info[TerminationState.CRASH_VEHICLE] or done_info[TerminationState.CRASH_BUILDING] or \
                   done_info[TerminationState.SUCCESS] or done_info[TerminationState.OUT_OF_ROAD]
            if not done_info[TerminationState.SUCCESS]:
                done = False

        if vehicle_id in self.stay_time_manager.entry_time and vehicle_id in self.stay_time_manager.exit_time:
            entry = self.stay_time_manager.entry_time[vehicle_id]
            exit = self.stay_time_manager.exit_time[vehicle_id]
            if (exit - entry) < self.config["vehicle_config"]["min_pass_steps"]:
                done = True
                done_info["out_of_road"] = True

        return done, done_info

    def get_single_observation(self):
        o = TollGateObservation(self.config)
        return o

    def step(self, actions):
        o, r, tm, tc, i = super(MultiAgentTollgateEnv, self).step(actions)
        self.stay_time_manager.record(self.agent_manager.active_agents, self.episode_step)
        return o, r, tm, tc, i

    def setup_engine(self):
        super(MultiAgentTollgateEnv, self).setup_engine()
        self.engine.update_manager("map_manager", MATollGatePGMapManager())


def _draw():
    env = MultiAgentTollgateEnv()
    o, _ = env.reset()
    from metadrive.utils.draw_top_down_map import draw_top_down_map
    import matplotlib.pyplot as plt

    plt.imshow(draw_top_down_map(env.current_map))
    plt.show()
    env.close()


def _expert():
    env = MultiAgentTollgateEnv(
        {
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 240,
                    "num_others": 4,
                    "distance": 50
                },
            },
            "save_level": 1.,
            "use_AI_protector": True,
            "debug_physics_world": True,

            # "use_render": True,
            "debug": True,
            "manual_control": True,
            "num_agents": 4,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step(env.action_space.sample())
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        tm.update({"total_r": total_r, "episode length": ep_s})
        # env.render(text=d)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis_debug_respawn():
    env = MultiAgentTollgateEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                "show_lidar": False,
            },
            "debug_physics_world": True,
            "use_render": True,
            "debug": False,
            "manual_control": True,
            "num_agents": 20,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        action = {k: [.0, 1.0] for k in env.vehicles.keys()}
        o, r, tm, tc, info = env.step(action)
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # d.update({"total_r": total_r, "episode length": ep_s})
        render_text = {
            "total_r": total_r,
            "episode length": ep_s,
            "cam_x": env.main_camera.camera_x,
            "cam_y": env.main_camera.camera_y,
            "cam_z": env.main_camera.top_down_camera_height
        }
        env.render(text=render_text)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis():
    env = MultiAgentTollgateEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                # "show_lidar": True,
                # "show_side_detector":True,
                # "show_lane_line_detector":True,
            },
            "traffic_density": 0.,
            "traffic_mode": "hybrid",
            "debug": True,
            "use_render": True,
            # "debug": True,
            "manual_control": True,
            "num_agents": 18,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step({k: [0, 1] for k in env.vehicles.keys()})
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # d.update({"total_r": total_r, "episode length": ep_s})
        render_text = {
            "total_r": total_r,
            "episode length": ep_s,
            "cam_x": env.main_camera.camera_x,
            "cam_y": env.main_camera.camera_y,
            "cam_z": env.main_camera.top_down_camera_height
        }
        track_v = env.agent_manager.object_to_agent(env.current_track_vehicle.name)
        if track_v in r:
            render_text["tack_v_reward"] = r[track_v]
        render_text["dist_to_right"] = env.current_track_vehicle.dist_to_right_side
        render_text["dist_to_left"] = env.current_track_vehicle.dist_to_left_side
        render_text["overspeed"] = env.current_track_vehicle.overspeed
        render_text["lane"] = env.current_track_vehicle.lane_index
        render_text["block"] = env.current_track_vehicle.navigation.current_road.block_ID()
        env.render(text=render_text)
        if tm["__all__"]:
            print(info)
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _profile():
    import time
    env = MultiAgentTollgateEnv({"num_agents": 8})
    obs, _ = env.reset()
    start = time.time()
    for s in range(10000):
        o, r, tm, tc, i = env.step(env.action_space.sample())

        # mask_ratio = env.engine.detector_mask.get_mask_ratio()
        # print("Mask ratio: ", mask_ratio)

        if all(tm.values()):
            env.reset()
        if (s + 1) % 100 == 0:
            print(
                "Finish {}/10000 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
                    s + 1,
                    time.time() - start, (s + 1) / (time.time() - start)
                )
            )
    print(f"(MetaDriveEnv) Total Time Elapsed: {time.time() - start}")


def _long_run():
    # Please refer to test_ma_Toll_reward_done_alignment()
    _out_of_road_penalty = 3
    env = MultiAgentTollgateEnv(
        {
            "num_agents": 8,
            "vehicle_config": {
                "lidar": {
                    "num_others": 8
                }
            },
            **dict(
                out_of_road_penalty=_out_of_road_penalty,
                crash_vehicle_penalty=1.333,
                crash_object_penalty=11,
                crash_vehicle_cost=13,
                crash_object_cost=17,
                out_of_road_cost=19,
            )
        }
    )
    try:
        obs, _ = env.reset()
        assert env.observation_space.contains(obs)
        for step in range(10000):
            act = env.action_space.sample()
            o, r, tm, tc, i = env.step(act)
            if step == 0:
                assert not any(tm.values())

            if any(tm.values()):
                print("Current Done: {}\nReward: {}".format(tm, r))
                for kkk, ddd in tm.items():
                    if ddd and kkk != "__all__":
                        print("Info {}: {}\n".format(kkk, i[kkk]))
                print("\n")

            for kkk, rrr in r.items():
                if rrr == -_out_of_road_penalty:
                    assert tm[kkk]

            if (step + 1) % 200 == 0:
                print(
                    "{}/{} Agents: {} {}\nO: {}\nR: {}\nD: {}\nI: {}\n\n".format(
                        step + 1, 10000, len(env.vehicles), list(env.vehicles.keys()),
                        {k: (oo.shape, oo.mean(), oo.min(), oo.max())
                         for k, oo in o.items()}, r, tm, i
                    )
                )
            if tm["__all__"]:
                print('Current step: ', step)
                break
    finally:
        env.close()


if __name__ == "__main__":
    # _draw()
    _vis()
    # _vis_debug_respawn()
    # _profile()
    # _long_run()
    # pygame_replay("tollgate", MultiAgentTollgateEnv)



import math

import numpy as np

from metadrive.utils.math import norm, get_vertical_vector


class InterpolatingLine:
    """
    This class provides point set with interpolating function
    """
    def __init__(self, points):
        points = np.asarray(points)[..., :2]
        self.segment_property, self._start_points, self._end_points = self._get_properties(points)
        self._distance_b_a = self._end_points - self._start_points
        self.length = sum([seg["length"] for seg in self.segment_property])

    def position(self, longitudinal: float, lateral: float) -> np.ndarray:
        return self.get_point(longitudinal, lateral)

    def local_coordinates(self, position, only_in_lane_point=False):
        """
        Finding the local coordinate of a given point when projected to this interpolating line.
        We will iterate over all segments, from start to end, and compute the relative position of the point
        w.r.t. the start point of the segment.
        Since each line segment is straight, the relative position of the position to the start point is exactly
        the local coordination.
        If the point is not fall into this line segment, then we will increment the "longitudinal position" by the
        length of this line segment, and then computing next line segment.

        Here we provide two implementations.

        Option 1: Iterate over all segments, and when finding the longitudinal position of the point w.r.t. the
        start point of the segment is negative, stop iteration and return current "accumulated longitudinal position"
        and lateral position.
        This option assumes the longitudinal position is accumulating increasingly when iterating over segments.
        Option 1 might be problematic in the case of extremely curved road.
        (PZH: But I can't image such case and why it fails option 1.)

        Option 2: Iterate over all segments and never stop the iteration until all segments are visited. Compute the
        "accumulated longitudinal position" and lateral position of the point in each segment when assuming the point
        is exactly falling into the segment. Then, search all reported coordinates in all segments, find the one
        with minimal lateral position, and report the value.
        This option is problematic as raised by PZH that the return value might be wrong.
        Let's say there is a correct segment A where the position is falling into, and there exists another segment
        B where the lateral position of the point in segment B is 0, but segment B is far away from A in longitudinal.
        In this case, the option 2 algorithm will return segment B because it has minimal lateral, but it is wrong.

        We will use Option 1.
        """
        min_dists = self.min_lineseg_dist(position, self._start_points, self._end_points, self._distance_b_a)
        target_segment_idx = np.argmin(min_dists)

        long = 0
        for idx, seg in enumerate(self.segment_property):
            if idx != target_segment_idx:
                long += seg["length"]
            else:
                delta_x = position[0] - seg["start_point"][0]
                delta_y = position[1] - seg["start_point"][1]
                long += delta_x * seg["direction"][0] + delta_y * seg["direction"][1]
                lateral = delta_x * seg["lateral_direction"][0] + delta_y * seg["lateral_direction"][1]
                return long, lateral

        # deprecated content
        # Four elements:
        #   accumulated longitude,
        #   segment-related longitude,
        #   segment-related lateral,
        # ret = []
        # exclude_ret = []
        # accumulate_len = 0
        #
        # # _debug = []
        #
        # for seg in self.segment_property:
        #     delta_x = position[0] - seg["start_point"][0]
        #     delta_y = position[1] - seg["start_point"][1]
        #     longitudinal = delta_x * seg["direction"][0] + delta_y * seg["direction"][1]
        #     lateral = delta_x * seg["lateral_direction"][0] + delta_y * seg["lateral_direction"][1]
        #     # _debug.append(longitudinal)
        #
        #     if longitudinal < 0.0:
        #         dist_square = norm(delta_x, delta_y)
        #         if dist_square < seg["length"] * 2:
        #             current_long = accumulate_len + longitudinal
        #             current_lat = lateral
        #             return current_long, current_lat
        #
        #     if not only_in_lane_point:
        #         ret.append([accumulate_len + longitudinal, longitudinal, lateral])
        #     else:
        #         if abs(lateral) <= self.width / 2 and -1. <= accumulate_len + longitudinal <= self.length + 1:
        #             ret.append([accumulate_len + longitudinal, longitudinal, lateral])
        #         else:
        #             exclude_ret.append([accumulate_len + longitudinal, longitudinal, lateral])
        #     accumulate_len += seg["length"]
        # if len(ret) == 0:
        #     # for corner case
        #     ret = exclude_ret
        # ret.sort(key=lambda seg: abs(seg[-1]))
        # return ret[0][0], ret[0][-1]

    def _get_properties(self, points):
        points = np.asarray(points)[..., :2]
        ret = []
        start_points = []
        end_points = []
        p_start_idx = 0
        while p_start_idx < len(points) - 1:
            for p_end_idx in range(p_start_idx + 1, len(points)):
                _p = points[p_start_idx] - points[p_end_idx]
                if norm(_p[0], _p[1]) > 1:
                    break
            p_start = points[p_start_idx]
            p_end = points[p_end_idx]

            _p = p_start - p_end
            if norm(_p[0], _p[1]) < 1e-6:
                p_start_idx = p_end_idx  # next
                continue

            seg_property = {
                "length": self.points_distance(p_start, p_end),
                "direction": np.asarray(self.points_direction(p_start, p_end)),
                "lateral_direction": np.asarray(self.points_lateral_direction(p_start, p_end)),
                "heading": self.points_heading(p_start, p_end),
                "start_point": p_start,
                "end_point": p_end
            }
            ret.append(seg_property)
            start_points.append(seg_property["start_point"])
            end_points.append(seg_property["end_point"])
            p_start_idx = p_end_idx  # next
        if len(ret) == 0:
            # static, length=zero
            seg_property = {
                "length": 0.1,
                "direction": np.asarray((1, 0)),
                "lateral_direction": np.asarray((0, 1)),
                "heading": 0,
                "start_point": points[0],
                "end_point": np.asarray([points[0][0] + 0.1, points[0][1]])
            }
            ret.append(seg_property)
            start_points.append(seg_property["start_point"])
            end_points.append(seg_property["end_point"])
        return ret, np.asarray(start_points), np.asarray(end_points)

    @staticmethod
    def points_distance(start_p, end_p):
        return norm((end_p - start_p)[0], (end_p - start_p)[1])

    @staticmethod
    def points_direction(start_p, end_p):
        return (end_p - start_p) / norm((end_p - start_p)[0], (end_p - start_p)[1])

    @staticmethod
    def points_lateral_direction(start_p, end_p):
        # direction = (end_p - start_p) / norm((end_p - start_p)[0], (end_p - start_p)[1])
        return np.asarray(get_vertical_vector(end_p - start_p)[1])

    @staticmethod
    def points_heading(start_p, end_p):
        return math.atan2(end_p[1] - start_p[1], end_p[0] - start_p[0])

    def get_point(self, longitudinal, lateral=None):
        """
        Get point on this line by interpolating
        """
        accumulate_len = 0
        for seg in self.segment_property:
            accumulate_len += seg["length"]
            if accumulate_len + 0.1 >= longitudinal:
                break
        if lateral is not None:
            return (seg["start_point"] + (longitudinal - accumulate_len + seg["length"]) *
                    seg["direction"]) + lateral * seg["lateral_direction"]
        else:
            return seg["start_point"] + (longitudinal - accumulate_len + seg["length"]) * seg["direction"]

    def get_heading_theta(self, longitudinal: float) -> float:
        """
        In rad
        """
        accumulate_len = 0

        assert len(self.segment_property) > 0

        for seg in self.segment_property:
            accumulate_len += seg["length"]
            if accumulate_len > longitudinal:
                return seg["heading"]

        return seg["heading"]

    def segment(self, longitudinal: float):
        """
        Return the segment piece on this lane of current position
        """
        accumulate_len = 0
        for index, seg in enumerate(self.segment_property):
            accumulate_len += seg["length"]
            if accumulate_len + 0.1 >= longitudinal:
                return self.segment_property[index]
        return self.segment_property[index]

    def lateral_direction(self, longitude):
        lane_segment = self.segment(longitude)
        lateral = lane_segment["lateral_direction"]
        return lateral

    def destroy(self):
        del self.segment_property
        self.segment_property = []
        self.length = None

    @staticmethod
    def min_lineseg_dist(p, a, b, d_ba=None):
        """Cartesian distance from point to line segment
        Edited to support arguments as series, from:
        https://stackoverflow.com/a/54442561/11208892

        Args:
            - p: np.array of single point, shape (2,) or 2D array, shape (x, 2)
            - a: np.array of shape (x, 2)
            - b: np.array of shape (x, 2)
        """
        # normalized tangent vectors
        p = np.asarray(p)
        if d_ba is None:
            d_ba = b - a
        d = np.divide(d_ba, (np.hypot(d_ba[:, 0], d_ba[:, 1]).reshape(-1, 1)))

        # signed parallel distance components
        # rowwise dot products of 2D vectors
        s = np.multiply(a - p, d).sum(axis=1)
        t = np.multiply(p - b, d).sum(axis=1)

        # clamped parallel distance
        h = np.maximum.reduce([s, t, np.zeros(len(s))])

        # perpendicular distance component
        # rowwise cross products of 2D vectors
        d_pa = p - a
        c = d_pa[:, 0] * d[:, 1] - d_pa[:, 1] * d[:, 0]
        min_dists = np.hypot(h, c)
        return min_dists


if __name__ == '__main__':
    test_p = [
        (611.5277397744468, 1829.6026943481352), (611.9289393579008, 1830.4997828926864),
        (612.330138941355, 1831.3968714372377), (612.7313385248092, 1832.293959981789),
        (613.1325381082634, 1833.1910485263402), (613.5337376917175, 1834.0881370708914),
        (613.9349372751717, 1834.985225615443), (614.3361368586259, 1835.8823141599942),
        (614.73733644208, 1836.7794027045454), (615.1385360255342, 1837.6764912490967),
        (615.5397446694475, 1838.573575740269), (615.9442295025767, 1839.4691858535757),
        (616.3546178958309, 1840.362106243068), (616.7708919643475, 1841.2522979949786),
        (617.1930335667643, 1842.1397223144547), (617.6210243060098, 1843.02434052725),
        (618.0548455301055, 1843.9061140814083), (618.494478332978, 1844.7850045489452),
        (618.9399035552839, 1845.660973627522), (619.3911017852438, 1846.533983142115),
        (619.8480533594889, 1847.4039950466793), (620.3107383639176, 1848.2709714258071),
        (620.3107383639176, 1848.2709714258071), (620.7713311311863, 1849.086333109443),
        (621.2652449066546, 1849.8819520853385), (621.7916348376419, 1850.656467426024),
        (622.3496005201381, 1851.4085543023612), (622.938187538968, 1852.1369262496978),
        (623.5563891003444, 1852.8403373683998), (624.2031477540157, 1853.5175844549963),
        (624.8773572020626, 1854.16750906029), (625.5778641912499, 1854.7889994709174),
        (626.3034704856974, 1855.3809926109623), (627.0430244723732, 1855.9555751328883),
        (627.7826528334156, 1856.5300621160332), (628.522281194458, 1857.1045490991783),
        (629.2619095555023, 1857.6790360823206)
    ]

    import matplotlib.pyplot as plt

    plt.gca().set_aspect('equal')

    def draw_polyline(polyline, colors, points, point_colors):
        for i in range(len(polyline) - 1):
            segment = [polyline[i], polyline[i + 1]]
            plt.plot(*zip(*segment), color=colors[i])

        for i, point in enumerate(points):
            plt.scatter(*point, color=point_colors[i])

        plt.show()

    line = InterpolatingLine(test_p)
    point = (619.8480533594889 + 10, 1847.4039950466793 - 4.5)
    dist = line.min_lineseg_dist(point, line._start_points, line._end_points, line._distance_b_a)
    min_seg = np.argmin(dist)

    # Example usage:
    polyline = np.concatenate([line._start_points, line._end_points[-1:]], axis=0)
    colors = ["blue"] * len(line._start_points)
    colors[min_seg] = "red"

    long, lat = line.local_coordinates(point)
    cross = line.position(long, 0)
    draw_polyline(
        polyline,
        colors=colors,
        points=[point, cross, line._end_points[min_seg], line._start_points[min_seg]],
        point_colors=["red", "green", "blue", "blue"]
    )



import copy

import numpy as np
from panda3d.bullet import BulletBoxShape, BulletGhostNode
from panda3d.core import Vec3

from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.map.pg_map import PGMap
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.parking_lot import ParkingLot
from metadrive.component.pgblock.t_intersection import TInterSection
from metadrive.component.road_network import Road
from metadrive.constants import MetaDriveType
from metadrive.constants import CollisionGroup
from metadrive.engine.engine_utils import get_engine
from metadrive.envs.marl_envs.multi_agent_metadrive import MultiAgentMetaDrive
from metadrive.manager.pg_map_manager import PGMapManager
from metadrive.utils import get_np_random, Config
from metadrive.utils.coordinates_shift import panda_vector, panda_heading
from metadrive.utils.pg.utils import rect_region_detection

MAParkingLotConfig = dict(
    in_spawn_roads=[
        Road(FirstPGBlock.NODE_2, FirstPGBlock.NODE_3),
        -Road(TInterSection.node(2, 0, 0), TInterSection.node(2, 0, 1)),
        -Road(TInterSection.node(2, 2, 0), TInterSection.node(2, 2, 1)),
    ],
    out_spawn_roads=None,  # auto fill
    spawn_roads=None,  # auto fill
    num_agents=10,
    parking_space_num=8,
    map_config=dict(exit_length=20, lane_num=1),
    top_down_camera_initial_x=80,
    top_down_camera_initial_y=0,
    top_down_camera_initial_z=120,
    vehicle_config={
        "enable_reverse": True,
        "random_navi_mark_color": True,
        "show_dest_mark": True,
        "show_navi_mark": False,
        "show_line_to_dest": True,
    },
)

from metadrive.manager.spawn_manager import SpawnManager


class ParkingLotSpawnManager(SpawnManager):
    """
    Manage parking spaces, when env.reset() is called, vehicles will be assigned to different spawn points including:
    parking space and entrances of parking lot, vehicle can not respawn in parking space which has been assigned to a
    vehicle who drives into this parking lot.
    """
    def __init__(self):
        super(ParkingLotSpawnManager, self).__init__()
        self.parking_space_available = set()
        self._parking_spaces = None
        self.v_dest_pair = {}

    def get_parking_space(self, v_id):
        if self._parking_spaces is None:
            self._parking_spaces = self.engine.map_manager.current_map.parking_space
            self.v_dest_pair = {}
            self.parking_space_available = set(copy.deepcopy(self._parking_spaces))
        assert len(self.parking_space_available) > 0
        parking_space_idx = self.np_random.choice([i for i in range(len(self.parking_space_available))])
        parking_space = list(self.parking_space_available)[parking_space_idx]
        self.parking_space_available.remove(parking_space)
        self.v_dest_pair[v_id] = parking_space
        return parking_space

    def after_vehicle_done(self, v_id):
        if v_id in self.v_dest_pair:
            dest = self.v_dest_pair.pop(v_id)
            self.parking_space_available.add(dest)

    def reset(self):
        self._parking_spaces = self.engine.map_manager.current_map.parking_space
        self.v_dest_pair = {}
        self.parking_space_available = set(copy.deepcopy(self._parking_spaces))
        super(ParkingLotSpawnManager, self).reset()

    def update_destination_for(self, vehicle_id, vehicle_config):
        # when agent re-joined to the game, call this to set the new route to destination
        end_roads = copy.deepcopy(self.engine.global_config["in_spawn_roads"])
        if Road(*vehicle_config["spawn_lane_index"][:-1]) in end_roads:
            end_road = self.engine.spawn_manager.get_parking_space(vehicle_id)
        else:
            end_road = -self.np_random.choice(end_roads)  # Use negative road!
        vehicle_config["destination"] = end_road.end_node
        return vehicle_config

    def get_available_respawn_places(self, map, randomize=False):
        """
        In each episode, we allow the vehicles to respawn at the start of road, randomize will give vehicles a random
        position in the respawn region
        """
        engine = get_engine()
        ret = {}
        for bid, bp in self.safe_spawn_places.items():
            if bid in self.spawn_places_used:
                continue
            if "P" not in bid and len(self.parking_space_available) == 0:
                # If no parking space, vehicles will never be spawned.
                continue
            # save time calculate once
            if not bp.get("spawn_point_position", False):
                lane = map.road_network.get_lane(bp["config"]["spawn_lane_index"])
                assert isinstance(lane, StraightLane), "Now we don't support respawn on circular lane"
                long = self.RESPAWN_REGION_LONGITUDE / 2
                spawn_point_position = lane.position(longitudinal=long, lateral=0)
                bp.force_update(
                    {
                        "spawn_point_heading": np.rad2deg(lane.heading_theta_at(long)),
                        "spawn_point_position": (spawn_point_position[0], spawn_point_position[1])
                    }
                )

            spawn_point_position = bp["spawn_point_position"]
            lane_heading = bp["spawn_point_heading"]
            result = rect_region_detection(
                engine, spawn_point_position, lane_heading, self.RESPAWN_REGION_LONGITUDE, self.RESPAWN_REGION_LATERAL,
                CollisionGroup.Vehicle
            )
            if (engine.global_config["debug"] or engine.global_config["debug_physics_world"]) \
                    and bp.get("need_debug", True):
                shape = BulletBoxShape(Vec3(self.RESPAWN_REGION_LONGITUDE / 2, self.RESPAWN_REGION_LATERAL / 2, 1))
                vis_body = engine.render.attach_new_node(BulletGhostNode("debug"))
                vis_body.node().addShape(shape)
                vis_body.setH(panda_heading(lane_heading))
                vis_body.setPos(panda_vector(spawn_point_position, z=2))
                engine.physics_world.dynamic_world.attach(vis_body.node())
                vis_body.node().setIntoCollideMask(CollisionGroup.AllOff)
                bp.force_set("need_debug", False)

            if not result.hasHit() or result.node.getName() != MetaDriveType.VEHICLE:
                new_bp = copy.deepcopy(bp).get_dict()
                if randomize:
                    new_bp["config"] = self._randomize_position_in_slot(new_bp["config"])
                ret[bid] = new_bp
                self.spawn_places_used.append(bid)
        return ret


class MAParkingLotMap(PGMap):
    def _generate(self):
        length = self.config["exit_length"]

        parent_node_path, physics_world = self.engine.worldNP, self.engine.physics_world
        assert len(self.road_network.graph) == 0, "These Map is not empty, please create a new map to read config"

        # Build a first-block
        last_block = FirstPGBlock(
            self.road_network,
            self.config[self.LANE_WIDTH],
            self.config[self.LANE_NUM],
            parent_node_path,
            physics_world,
            length=length
        )
        self.blocks.append(last_block)

        last_block = ParkingLot(1, last_block.get_socket(0), self.road_network, 1, ignore_intersection_checking=False)
        last_block.construct_block(
            parent_node_path, physics_world, {"one_side_vehicle_number": int(self.config["parking_space_num"] / 2)}
        )
        self.blocks.append(last_block)
        self.parking_space = last_block.dest_roads
        self.parking_lot = last_block

        # Build ParkingLot
        TInterSection.EXIT_PART_LENGTH = 10
        last_block = TInterSection(
            2, last_block.get_socket(index=0), self.road_network, random_seed=1, ignore_intersection_checking=False
        )
        last_block.construct_block(
            parent_node_path,
            physics_world,
            extra_config={
                "t_type": 1,
                "change_lane_num": 0
                # Note: lane_num is set in config.map_config.lane_num
            }
        )
        self.blocks.append(last_block)


class MAParkinglotPGMapManager(PGMapManager):
    def reset(self):
        config = self.engine.global_config
        if len(self.spawned_objects) == 0:
            _map = self.spawn_object(MAParkingLotMap, map_config=config["map_config"], random_seed=None)
        else:
            assert len(self.spawned_objects) == 1, "It is supposed to contain one map in this manager"
            _map = self.spawned_objects.values()[0]
        self.load_map(_map)
        self.current_map.spawn_roads = config["spawn_roads"]


class MultiAgentParkingLotEnv(MultiAgentMetaDrive):
    """
    Env will be done when vehicle is on yellow or white continuous lane line!
    """
    @staticmethod
    def default_config() -> Config:
        return MultiAgentMetaDrive.default_config().update(MAParkingLotConfig, allow_add_new_key=True)

    @staticmethod
    def _get_out_spawn_roads(parking_space_num):
        ret = []
        for i in range(1, parking_space_num + 1):
            ret.append(Road(ParkingLot.node(1, i, 5), ParkingLot.node(1, i, 6)))
        return ret

    def _post_process_config(self, config):
        ret_config = super(MultiAgentParkingLotEnv, self)._post_process_config(config)
        # add extra assert
        parking_space_num = ret_config["parking_space_num"]
        assert parking_space_num % 2 == 0, "number of parking spaces must be multiples of 2"
        assert parking_space_num >= 4, "minimal number of parking space is 4"
        ret_config["out_spawn_roads"] = self._get_out_spawn_roads(parking_space_num)
        ret_config["spawn_roads"] = ret_config["in_spawn_roads"] + ret_config["out_spawn_roads"]
        ret_config["map_config"]["parking_space_num"] = ret_config["parking_space_num"]
        return ret_config

    def _respawn_single_vehicle(self, randomize_position=False):
        """
        Exclude destination parking space
        """
        safe_places_dict = self.engine.spawn_manager.get_available_respawn_places(
            self.current_map, randomize=randomize_position
        )
        # ===== filter spawn places =====
        filter_ret = {}
        for id, config in safe_places_dict.items():
            spawn_l_index = config["config"]["spawn_lane_index"]
            spawn_road = Road(spawn_l_index[0], spawn_l_index[1])
            if spawn_road in self.config["in_spawn_roads"]:
                if len(self.engine.spawn_manager.parking_space_available) > 0:
                    filter_ret[id] = config
            else:
                # spawn in parking space
                if ParkingLot.is_in_direction_parking_space(spawn_road):
                    # avoid sweep test bug
                    spawn_road = self.current_map.parking_lot.out_direction_parking_space(spawn_road)
                    config["config"]["spawn_lane_index"] = (spawn_road.start_node, spawn_road.end_node, 0)
                if spawn_road in self.engine.spawn_manager.parking_space_available:
                    # not other vehicle's destination
                    filter_ret[id] = config

        # ===== same as super() =====
        if len(safe_places_dict) == 0:
            return None, None, None
        born_place_index = get_np_random(self._DEBUG_RANDOM_SEED).choice(list(safe_places_dict.keys()), 1)[0]
        new_spawn_place = safe_places_dict[born_place_index]

        new_agent_id, vehicle, step_info = self.agent_manager.propose_new_vehicle()
        new_spawn_place_config = new_spawn_place["config"]
        new_spawn_place_config = self.engine.spawn_manager.update_destination_for(new_agent_id, new_spawn_place_config)
        vehicle.config.update(new_spawn_place_config)
        vehicle.reset()
        after_step_info = vehicle.after_step()
        step_info.update(after_step_info)
        self.dones[new_agent_id] = False  # Put it in the internal dead-tracking dict.

        new_obs = self.observations[new_agent_id].observe(vehicle)
        return new_agent_id, new_obs, step_info

    def done_function(self, vehicle_id):
        done, info = super(MultiAgentParkingLotEnv, self).done_function(vehicle_id)
        if done:
            self.engine.spawn_manager.after_vehicle_done(vehicle_id)
        return done, info

    def _is_out_of_road(self, vehicle):
        # A specified function to determine whether this vehicle should be done.
        return vehicle.on_yellow_continuous_line or (not vehicle.on_lane) or vehicle.crash_sidewalk
        # ret = vehicle.out_of_route
        # return ret

    def setup_engine(self):
        super(MultiAgentParkingLotEnv, self).setup_engine()
        self.engine.update_manager("spawn_manager", ParkingLotSpawnManager())
        self.engine.update_manager("map_manager", MAParkinglotPGMapManager())


def _draw():
    env = MultiAgentParkingLotEnv()
    o, _ = env.reset()
    from metadrive.utils.draw_top_down_map import draw_top_down_map
    import matplotlib.pyplot as plt

    plt.imshow(draw_top_down_map(env.current_map))
    plt.show()
    env.close()


def _expert():
    env = MultiAgentParkingLotEnv(
        {
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 240,
                    "num_others": 4,
                    "distance": 50
                },
            },
            "save_level": 1.,
            "use_AI_protector": True,
            "debug_physics_world": True,

            # "use_render": True,
            "debug": True,
            "manual_control": True,
            "num_agents": 3,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step(env.action_space.sample())
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        tm.update({"total_r": total_r, "episode length": ep_s})
        # env.render(text=d)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis_debug_respawn():
    env = MultiAgentParkingLotEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                "show_lidar": False,
            },
            "debug_physics_world": True,
            "use_render": True,
            "debug": False,
            "manual_control": True,
            "num_agents": 11,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        action = {k: [0.0, .0] for k in env.vehicles.keys()}
        o, r, tm, tc, info = env.step(action)
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # d.update({"total_r": total_r, "episode length": ep_s})
        render_text = {
            "total_r": total_r,
            "episode length": ep_s,
            "cam_x": env.main_camera.camera_x,
            "cam_y": env.main_camera.camera_y,
            "cam_z": env.main_camera.top_down_camera_height
        }
        env.render(text=render_text)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis():
    # vis_big(block_type_version="v2")
    env = MultiAgentParkingLotEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                "show_lidar": False,
            },
            "debug_static_world": True,
            "debug_physics_world": True,
            "global_light": True,
            "use_render": True,
            "debug": True,
            "manual_control": True,
            "num_agents": 7,
            "delay_done": 10,
            # "parking_space_num": 4
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        actions = {k: [1.0, .0] for k in env.vehicles.keys()}
        if len(env.vehicles) == 1:
            actions = {k: [-1.0, .0] for k in env.vehicles.keys()}
        o, r, tm, tc, info = env.step(actions)
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # d.update({"total_r": total_r, "episode length": ep_s})
        if len(env.vehicles) != 0:
            v = env.current_track_vehicle
            dist = v.dist_to_left_side, v.dist_to_right_side
            ckpt_idx = v.navigation._target_checkpoints_index
        else:
            dist = (0, 0)
            ckpt_idx = (0, 0)

        render_text = {
            "total_r": total_r,
            "episode length": ep_s,
            "cam_x": env.main_camera.camera_x,
            "cam_y": env.main_camera.camera_y,
            "cam_z": env.main_camera.top_down_camera_height,
            "alive": len(env.vehicles),
            "dist_right_left": dist,
            "ckpt_idx": ckpt_idx,
            "parking_space_num": len(env.engine.spawn_manager.parking_space_available)
        }
        if len(env.vehicles) > 0:
            v = env.current_track_vehicle
            # print(v.navigation.checkpoints)
            render_text["current_road"] = v.navigation.current_road

        env.render(text=render_text)
        for kkk, ddd in d.items():
            if ddd and kkk != "__all__":
                print(
                    "{} done! State: {}".format(
                        kkk, {
                            "arrive_dest": info[kkk]["arrive_dest"],
                            "out_of_road": info[kkk]["out_of_road"],
                            "crash": info[kkk]["crash"],
                            "max_step": info[kkk]["max_step"],
                        }
                    )
                )
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            env.reset()
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _profile():
    import time
    env = MultiAgentParkingLotEnv({"num_agents": 10})
    obs, _ = env.reset()
    start = time.time()
    for s in range(10000):
        o, r, tm, tc, i = env.step(env.action_space.sample())

        # mask_ratio = env.engine.detector_mask.get_mask_ratio()
        # print("Mask ratio: ", mask_ratio)

        if all(tm.values()):
            env.reset()
        if (s + 1) % 100 == 0:
            print(
                "Finish {}/10000 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
                    s + 1,
                    time.time() - start, (s + 1) / (time.time() - start)
                )
            )
    print(f"(MAParkingLot) Total Time Elapse: {time.time() - start}")


def _long_run():
    # Please refer to test_ma_ParkingLot_reward_done_alignment()
    _out_of_road_penalty = 3
    env = MultiAgentParkingLotEnv(
        {
            "num_agents": 3,
            "vehicle_config": {
                "lidar": {
                    "num_others": 8
                }
            },
            **dict(
                out_of_road_penalty=_out_of_road_penalty,
                crash_vehicle_penalty=1.333,
                crash_object_penalty=11,
                crash_vehicle_cost=13,
                crash_object_cost=17,
                out_of_road_cost=19,
            )
        }
    )
    try:
        obs, _ = env.reset()
        assert env.observation_space.contains(obs)
        for step in range(10000):
            act = env.action_space.sample()
            o, r, tm, tc, i = env.step(act)
            if step == 0:
                assert not any(d.values())

            if any(tm.values()):
                print("Current Done: {}\nReward: {}".format(d, r))
                for kkk, ddd in tm.items():
                    if ddd and kkk != "__all__":
                        print("Info {}: {}\n".format(kkk, i[kkk]))
                print("\n")

            for kkk, rrr in r.items():
                if rrr == -_out_of_road_penalty:
                    assert d[kkk]

            if (step + 1) % 200 == 0:
                print(
                    "{}/{} Agents: {} {}\nO: {}\nR: {}\nD: {}\nI: {}\n\n".format(
                        step + 1, 10000, len(env.vehicles), list(env.vehicles.keys()),
                        {k: (oo.shape, oo.mean(), oo.min(), oo.max())
                         for k, oo in o.items()}, r, d, i
                    )
                )
            if tm["__all__"]:
                print('Current step: ', step)
                break
    finally:
        env.close()


if __name__ == "__main__":
    # _draw()
    _vis()
    # _vis_debug_respawn()
    # _profile()
    # _long_run()
    # pygame_replay("parking", MultiAgentParkingLotEnv, False, other_traj="metasvodist_parking_best.json")
    # panda_replay(
    #     "parking",
    #     MultiAgentParkingLotEnv,
    #     False,
    #     other_traj="metasvodist_parking_best.json",
    #     extra_config={"global_light": True}
    # )



import gymnasium as gym
import numpy as np

from metadrive.obs.state_obs import LidarStateObservation
from metadrive.utils import clip


class NuPlanObservation(LidarStateObservation):
    MAX_LATERAL_DIST = 20

    def __init__(self, *args, **kwargs):
        super(NuPlanObservation, self).__init__(*args, **kwargs)
        self.lateral_dist = 0

    @property
    def observation_space(self):
        shape = list(self.state_obs.observation_space.shape)
        if self.config["vehicle_config"]["lidar"]["num_lasers"] > 0 and self.config["vehicle_config"]["lidar"][
                "distance"] > 0:
            # Number of lidar rays and distance should be positive!
            lidar_dim = self.config["vehicle_config"]["lidar"][
                "num_lasers"] + self.config["vehicle_config"]["lidar"]["num_others"] * 4
            if self.config["vehicle_config"]["lidar"]["add_others_navi"]:
                lidar_dim += self.config["vehicle_config"]["lidar"]["num_others"] * 4
            shape[0] += lidar_dim
        shape[0] += 1  # add one dim for sensing lateral distance to the sdc trajectory
        return gym.spaces.Box(-0.0, 1.0, shape=tuple(shape), dtype=np.float32)

    def state_observe(self, vehicle):
        ret = super(NuPlanObservation, self).state_observe(vehicle)
        lateral_obs = self.lateral_dist / self.MAX_LATERAL_DIST
        return np.concatenate([ret, [clip((lateral_obs + 1) / 2, 0.0, 1.0)]])

    def reset(self, env, vehicle=None):
        super(NuPlanObservation, self).reset(env, vehicle)
        self.lateral_dist = 0



import time

import numpy as np

from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.envs import MetaDriveEnv


def get_result(env):
    obs, _ = env.reset()
    start = time.time()
    max_speed_km_h = 0.0
    reported_max_speed = None
    reported_end = None
    reported_start = None
    reported_rotation = None
    start_heading = env.vehicle.heading_theta
    rotate_start_pos = None
    max_speed_loc = None
    for s in range(10000):
        if s < 20:
            action = np.array([0.0, 0.0])
        elif env.vehicle.speed_km_h < 100 and not reported_max_speed:
            action = np.array([0.0, 1.0])
        else:
            action = np.array([0.0, -1.0])
            # action = np.array([0.0, 0.0])

        if s > 20 and env.vehicle.speed_km_h > 1.0 and not reported_start:
            # print("Start the car at {}".format(s))
            reported_start = s
            start_time = time.time()

        if s > 20 and env.vehicle.speed_km_h >= 100 and not reported_max_speed:
            spend = (s - 1 - reported_start) * 0.1
            print(
                "Achieve max speed: {} at {}. Spend {} s. Current location: {}".format(
                    max_speed_km_h, s - 1, spend, env.vehicle.position
                )
            )
            # print("real time spend to acc: {}".format(time.time() - start_time))
            reported_max_speed = s
            max_speed_loc = env.vehicle.position

        max_speed_km_h = max(max_speed_km_h, env.vehicle.speed_km_h)

        if s > 20 and env.vehicle.speed_km_h <= 1.0 and reported_max_speed and not reported_end:
            dist = env.vehicle.position - max_speed_loc
            dist = dist[0]
            # print("Stop the car at {}. Distance {}. Current location: {}".format(s, dist, env.vehicle.position))
            reported_end = True

        speed = env.vehicle.speed_km_h
        current_heading = env.vehicle.heading_theta
        if reported_end and not reported_rotation:
            if rotate_start_pos is None:
                rotate_start_pos = env.vehicle.position
            if speed < 99:
                action = np.array([0.0, 1.0])
            else:
                action = np.array([-1.0, 1.0])
                if abs(current_heading - start_heading) >= np.pi / 2:
                    rotate_displacement = np.asarray(env.vehicle.position) - np.asarray(rotate_start_pos)
                    reported_rotation = True

        o, r, tm, tc, i = env.step(action)

        if reported_max_speed and reported_start and reported_end and reported_rotation:
            break
        # env.render()
    return spend, dist, rotate_displacement


if __name__ == '__main__':

    for friction in [0.8]:
        env = MetaDriveEnv(
            {
                "num_scenarios": 1,
                "traffic_density": 0.0,
                "start_seed": 4,
                "manual_control": True,
                "use_render": True,
                "image_observation": True,
                "vehicle_config": {
                    "max_engine_force": 1000,
                    "max_brake_force": 100,
                    "max_steering": 40,
                    "wheel_friction": friction
                },
                "map_config": {
                    BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_SEQUENCE,
                    BaseMap.GENERATE_CONFIG: "SSSSSSSSSS"
                }
            }
        )
        acc_time, brake_dist, rotate_dis = get_result(env)
        print(
            "Friction {}. Acceleration time: {:.3f}. Brake distance: {:.3f}. Rotation 90 degree displacement X: {:.3f}, Y: {:.3f}\n\n"
            .format(friction, acc_time, brake_dist, rotate_dis[0], rotate_dis[1])
        )
        env.close()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error

import math
from panda3d.core import PNMImage


class IESDataset(object):
    """ Please refer to the native C++ implementation for docstrings and comments.
    This is just the python implementation, which does not contain documentation! """
    def __init__(self):
        self._vertical_angles = None
        self._horizontal_angles = None
        self._candela_values = None

    def set_vertical_angles(self, vertical_angles):
        self._vertical_angles = vertical_angles

    def set_horizontal_angles(self, horizontal_angles):
        self._horizontal_angles = horizontal_angles

    def set_candela_values(self, candela_values):
        self._candela_values = candela_values

    def generate_dataset_texture_into(self, dest_tex, layer_index):
        resolution_vertical = dest_tex.get_y_size()
        resolution_horizontal = dest_tex.get_x_size()

        dest = PNMImage(resolution_vertical, resolution_horizontal, 1, 65535)

        for vert in range(resolution_vertical):
            for horiz in range(resolution_horizontal):
                vert_angle = vert / (resolution_vertical - 1.0)
                vert_angle = math.cos(vert_angle * math.pi) * 90.0 + 90.0
                horiz_angle = horiz / (resolution_horizontal - 1.0) * 360.0
                candela = self.get_candela_value(vert_angle, horiz_angle)
                dest.set_xel(vert, horiz, candela)

        dest_tex.load(dest, layer_index, 0)

    def get_candela_value(self, vertical_angle, horizontal_angle):  # noqa # pylint: disable=unused-argument
        # NOTICE: Since python is slower, we always only assume a dataset without
        # horizontal angles. This still produces convincing results, but does
        # generate much faster.
        return self.get_vertical_candela_value(0, vertical_angle)

    def get_candela_value_from_index(self, vertical_angle_idx, horizontal_angle_idx):
        index = vertical_angle_idx + horizontal_angle_idx * len(self._vertical_angles)
        return self._candela_values[index]

    def get_vertical_candela_value(self, horizontal_angle_idx, vertical_angle):
        if vertical_angle < 0.0:
            return 0.0

        if vertical_angle > self._vertical_angles[len(self._vertical_angles) - 1]:
            return 0.0

        for vertical_index in range(1, len(self._vertical_angles)):
            curr_angle = self._vertical_angles[vertical_index]

            if curr_angle > vertical_angle:
                prev_angle = self._vertical_angles[vertical_index - 1]
                prev_value = self.get_candela_value_from_index(vertical_index - 1, horizontal_angle_idx)
                curr_value = self.get_candela_value_from_index(vertical_index, horizontal_angle_idx)
                lerp = (vertical_angle - prev_angle) / (curr_angle - prev_angle)
                assert lerp >= 0.0 and lerp <= 1.0
                return curr_value * lerp + prev_value * (1.0 - lerp)
        return 0.0



import os
import signal
import sys

from tqdm import tqdm

from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.policy.idm_policy import WaymoIDMPolicy

try:
    from metadrive.utils.waymo.waymo_type import WaymoAgentType
    from metadrive.utils.waymo.waymo_type import WaymoRoadLineType, WaymoRoadEdgeType
finally:
    pass

import logging

logger = logging.getLogger(__name__)

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)


def handler(signum, frame):
    raise Exception("end of time")


if __name__ == "__main__":
    scenario_data_path = sys.argv[1]
    start = int(sys.argv[2])
    processed_data_path = scenario_data_path + "_filtered"
    if not os.path.exists(processed_data_path):
        os.mkdir(processed_data_path)
    if not os.path.exists(scenario_data_path) or not os.path.exists(processed_data_path):
        raise ValueError("Path Not exist")
    num_scenarios = len(os.listdir(scenario_data_path))
    max_step = 1500
    min_step = 50

    env = WaymoEnv(
        {
            "use_render": False,
            "agent_policy": WaymoIDMPolicy,
            "data_directory": scenario_data_path,
            "start_scenario_index": start * 1000,
            "num_scenarios": num_scenarios,
            "store_map": False,
            # "manual_control": True,
            # "debug":True,
            "no_traffic": True,
            "horizon": 1500,
        }
    )
    try:
        env.reset()
    except:
        pass
    finally:
        pass
    for i in tqdm(range(num_scenarios)):
        try:
            signal.signal(signal.SIGALRM, handler)
            signal.alarm(10)
            env.reset(seed=i)
            while True:
                o, r, tm, tc, info = env.step([0, 0])
                if tm or tc or env.episode_step > max_step:
                    if info["arrive_dest"] and env.episode_step > min_step:
                        os.rename(
                            os.path.join(scenario_data_path, "{}.pkl".format(i + start * 1000)),
                            os.path.join(processed_data_path, "{}.pkl".format(i + start * 1000))
                        )
                    break
        except:
            # print("\n No Route or Timeout, Fail, Seed: {}".format(i))
            pass



# Author: Epihaius
# Date: 2019-06-21
#
# This module contains a class to generate box primitives.

from .base import *


class BoxMaker(ModelMaker):
    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, pos):
        self._center = pos

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width):
        self._width = width

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, depth):
        self._depth = depth

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height

    @property
    def open_sides(self):
        return self._open_sides

    @open_sides.setter
    def open_sides(self, side_ids):
        self._open_sides = side_ids

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, thickness):
        self._thickness = thickness

    def __init__(
        self,
        center=None,
        width=1.,
        depth=1.,
        height=1.,
        segments=None,
        open_sides=(),
        thickness=None,
        inverted=False,
        vertex_color=None,
        has_uvs=True,
        tex_units=None,
        tex_offset=None,
        tex_rotation=None,
        tex_scale=None
    ):
        """
        This class generates box model primitives with the given parameters:

            center (sequence or None):
                the position of the center of the geometry in object space;
                default is at the origin (the object-space (0., 0., 0.) point);
            width (float):
                dimension along the X-axis; cannot be negative;
                default = 1.;
            depth (float):
                dimension along the Y-axis; cannot be negative;
                default = 1.;
            height (float):
                dimension along the Z-axis; cannot be negative;
                default = 1.;
            segments (dict of ints):
                the number of side subdivisions in each dimension:
                    "width": minimum = 1, default = 1;
                    "depth": minimum = 1, default = 1;
                    "height": minimum = 1, default = 1;
            open_sides (sequence of strings):
                any combination of "left", "right", "back", "front", "bottom", "top";
                sides of the box that should not be created;
                default = empty sequence (no open sides);
            thickness (float or None):
                offset of inner box sides (in conjunction with open_sides, can lead
                to a number of different shapes, other than an actual box);
                cannot be negative;
                default = None (no inner box).

        The parameters common to all primitive types are documented in the
        ModelMaker base class.

        The surfaces of this primitive type can be referred to in the relevant
        parameters and properties as follows:
            "left", "right", "back", "front", "bottom",
            "top", "inner_left", "inner_right", "inner_back",
            "inner_front", "inner_bottom", "inner_top".

        Ranges of vertex indices of each side can be retrieved through the
        vertex_ranges property, as a dict of (start_index, end_index) tuples
        (empty tuple if the side was not created), with end_index not
        included in the range.

        """

        surface_ids = (
            "left", "right", "back", "front", "bottom", "top", "inner_left", "inner_right", "inner_back", "inner_front",
            "inner_bottom", "inner_top"
        )

        ModelMaker.__init__(
            self, segments, inverted, vertex_color, has_uvs, tex_units, tex_offset, tex_rotation, tex_scale, surface_ids
        )

        self._center = center
        self._width = width
        self._depth = depth
        self._height = height
        self._open_sides = open_sides
        self._thickness = thickness
        self._side_ids = {"left": "-yz", "right": "yz", "back": "-zx", "front": "zx", "bottom": "-xy", "top": "xy"}
        self._side_names = {"-yz": "left", "yz": "right", "-zx": "back", "zx": "front", "-xy": "bottom", "xy": "top"}

    def reset(self):

        ModelMaker.reset(self)

        self._center = None
        self._width = 1.
        self._depth = 1.
        self._height = 1.
        self._open_sides = ()
        self._thickness = None

    def __define_quads(self, indices, index_offset, direction, segs1, segs2=1):

        for i in range(segs2):

            for j in range(segs1):

                vi1 = index_offset + i * (segs1 + 1) + j
                vi2 = vi1 + 1
                vi3 = vi2 + segs1
                vi4 = vi3 + 1

                if self._inverted:
                    inds = (vi1, vi4, vi2) if direction == 1 else (vi1, vi2, vi4)
                else:
                    inds = (vi1, vi2, vi4) if direction == 1 else (vi1, vi4, vi2)

                indices.extend(inds)

                if self._inverted:
                    inds = (vi1, vi3, vi4) if direction == 1 else (vi1, vi4, vi3)
                else:
                    inds = (vi1, vi4, vi3) if direction == 1 else (vi1, vi3, vi4)

                indices.extend(inds)

    def __define_vertices(
        self, values, dims, dim1, dim2, c1, c2, sign, plane_id, direction, segs, pos, offset1, offset2, normal,
        axis1_index, axis2_index, primary_index, indices
    ):

        stride = 8 if self._has_uvs else 6  # number of floats on each vertex data row
        index_offset = len(values) // stride

        if primary_index == 1:
            t = ((dim2, c1, c2), (dim1, 0., 0.))
            index1 = axis1_index
            index2 = axis2_index
            offs1 = offset1
            offs2 = offset2
        else:
            t = ((dim1, 0., 0.), (dim2, c2, c1))
            index1 = axis2_index
            index2 = axis1_index
            offs1 = offset2
            offs2 = offset1

        if self._has_uvs:

            side_id = "-" if direction == -1 else ""
            side_id += plane_id
            side_name = self._side_names[side_id]
            tex_size = None

            if self._tex_units and side_name in self._tex_units:
                tex_w, tex_h = self._tex_units[side_name]
                tex_size = (tex_h, tex_w) if plane_id == "zx" else (tex_w, tex_h)

            mat = self._get_tex_xform(side_name)

        for dim, start1, start2 in t:

            if sign == "-":
                coord = start1 - dims[index1] * .5
            else:
                coord = dims[index1] * .5 - start1

            pos[index1] = coord + offs1

            a = coord / dims[index1] + .5

            for i in range(segs + 1):

                j = i / segs

                if sign == "-":
                    coord = start2 - dims[index2] * .5 + j * dim
                else:
                    coord = dims[index2] * .5 - j * dim - start2

                pos[index2] = coord + offs2

                values.extend(pos)
                values.extend(normal)

                if self._has_uvs:

                    b = coord / dims[index2] + .5

                    if primary_index == 1:
                        u = (-b if plane_id == "zx" else a) * direction
                    else:
                        u = (-a if plane_id == "zx" else b) * direction

                    if (direction > 0 if plane_id == "zx" else direction < 0):
                        u += 1.

                    if self._inverted:
                        u = 1. - u

                    if primary_index == 1:
                        v = a if plane_id == "zx" else b
                    else:
                        v = b if plane_id == "zx" else a

                    if tex_size:
                        if plane_id == "zx":
                            u *= dims[axis2_index] / tex_size[1]
                            v *= dims[axis1_index] / tex_size[0]
                        else:
                            u *= dims[axis1_index] / tex_size[0]
                            v *= dims[axis2_index] / tex_size[1]

                    if mat:
                        u, v = mat.xform_point(Point2(u, v))

                    values.extend((u, v))

        self.__define_quads(indices, index_offset, direction, segs)

    def generate(self):

        center = (0., 0., 0.) if self._center is None else self._center
        width = max(.001, self._width)
        depth = max(.001, self._depth)
        height = max(.001, self._height)
        dims = (width, depth, height)
        side_ids = self._side_ids
        side_names = self._side_names
        open_side_ids = [side_ids[name] for name in self._open_sides]
        thickness = self._thickness

        if thickness is not None:

            # Set limits for the given thickness and derive the dimensions and
            # corner coordinates of the inner box

            th = max(0., thickness)
            inner_corners = {}
            inner_dims = {}

            th_left = 0. if "-yz" in open_side_ids else min(width, th)
            th_right = 0. if "yz" in open_side_ids else min(width, th)

            if th_left + th_right > width:
                th_left = th_right = width * .5

            inner_corners["-x"] = th_left
            inner_corners["x"] = th_right
            inner_dims["x"] = width - th_left - th_right

            th_back = 0. if "-zx" in open_side_ids else min(depth, th)
            th_front = 0. if "zx" in open_side_ids else min(depth, th)

            if th_back + th_front > depth:
                th_back = th_front = depth * .5

            inner_corners["-y"] = th_back
            inner_corners["y"] = th_front
            inner_dims["y"] = depth - th_back - th_front

            th_bottom = 0. if "-xy" in open_side_ids else min(height, th)
            th_top = 0. if "xy" in open_side_ids else min(height, th)

            if th_bottom + th_top > height:
                th_bottom = th_top = height * .5

            inner_corners["-z"] = th_bottom
            inner_corners["z"] = th_top
            inner_dims["z"] = height - th_bottom - th_top

        segs = {} if self._segments is None else self._segments
        segs_w = max(1, segs.get("width", 1))
        segs_d = max(1, segs.get("depth", 1))
        segs_h = max(1, segs.get("height", 1))
        segs = {"x": segs_w, "y": segs_d, "z": segs_h}
        inverted = self._inverted
        has_uvs = self._has_uvs
        tex_units = self._tex_units
        tex_offset = self._tex_offset
        tex_rotation = self._tex_rotation
        tex_scale = self._tex_scale
        self._vert_ranges = vert_ranges = {
            "left": (),
            "right": (),
            "back": (),
            "front": (),
            "bottom": (),
            "top": (),
            "inner_left": (),
            "inner_right": (),
            "inner_back": (),
            "inner_front": (),
            "inner_bottom": (),
            "inner_top": ()
        }
        stride = 8 if has_uvs else 6  # number of floats on each vertex data row
        values = array.array("f", [])
        indices = array.array("I", [])
        pos = Point3()

        # Define the vertices for each side of the box

        for (axis1_id, axis2_id, axis3_id) in ("xyz", "zxy", "yzx"):

            plane_id = axis1_id + axis2_id
            axis1_index = "xyz".index(axis1_id)
            axis2_index = "xyz".index(axis2_id)
            axis3_index = "xyz".index(axis3_id)
            offset1 = center[axis1_index]
            offset2 = center[axis2_index]
            offset3 = center[axis3_index]
            segs1 = segs[axis1_id]
            segs2 = segs[axis2_id]
            segs3 = segs[axis3_id]

            for direction in (-1, 1):

                side_id = "-" if direction == -1 else ""
                side_id += plane_id

                normal = Vec3()
                normal[axis3_index] = direction * (-1. if inverted else 1.)

                if side_id in open_side_ids:

                    if thickness is not None:

                        pos[axis3_index] = .5 * dims[axis3_index] * direction + offset3
                        index_offset = len(values) // stride

                        for sign in ("-", ""):

                            c1 = inner_corners[sign + axis1_id]
                            c2 = inner_corners[sign + axis2_id]

                            if c1 > 0.:
                                dim1 = dims[axis2_index]
                                dim2 = inner_dims[axis2_id]
                                self.__define_vertices(
                                    values, dims, dim1, dim2, c1, c2, sign, plane_id, direction, segs2, pos, offset1,
                                    offset2, normal, axis1_index, axis2_index, 1, indices
                                )

                            if c2 > 0.:
                                dim1 = dims[axis1_index]
                                dim2 = inner_dims[axis1_id]
                                self.__define_vertices(
                                    values, dims, dim1, dim2, c1, c2, sign, plane_id, direction, segs1, pos, offset1,
                                    offset2, normal, axis1_index, axis2_index, 2, indices
                                )

                        side_name = side_names[side_id]
                        end = len(values) // stride
                        vert_ranges[side_name] = (index_offset, end)

                else:

                    pos[axis3_index] = .5 * dims[axis3_index] * direction + offset3
                    index_offset = len(values) // stride
                    side_name = side_names[side_id]

                    if has_uvs:

                        tex_size = None

                        if tex_units and side_name in tex_units:
                            tex_w, tex_h = tex_units[side_name]
                            tex_size = (tex_h, tex_w) if plane_id == "zx" else (tex_w, tex_h)

                        mat = self._get_tex_xform(side_name)

                    for i in range(segs2 + 1):

                        b = i / segs2
                        pos[axis2_index] = (-.5 + b) * dims[axis2_index] + offset2

                        for j in range(segs1 + 1):

                            a = j / segs1
                            pos[axis1_index] = (-.5 + a) * dims[axis1_index] + offset1

                            values.extend(pos)
                            values.extend(normal)

                            if has_uvs:

                                u = (-b if plane_id == "zx" else a) * direction
                                u += (1. if (direction > 0 if plane_id == "zx" else direction < 0) else 0.)

                                if inverted:
                                    u = 1. - u

                                v = a if plane_id == "zx" else b

                                if tex_size:
                                    if plane_id == "zx":
                                        u *= dims[axis2_index] / tex_size[1]
                                        v *= dims[axis1_index] / tex_size[0]
                                    else:
                                        u *= dims[axis1_index] / tex_size[0]
                                        v *= dims[axis2_index] / tex_size[1]

                                if mat:
                                    u, v = mat.xform_point(Point2(u, v))

                                values.extend((u, v))

                    self.__define_quads(indices, index_offset, direction, segs1, segs2)

                    end = len(values) // stride
                    vert_ranges[side_name] = (index_offset, end)

        # Create the geometry structures

        if thickness is not None:

            # If a thickness is given, an inner box needs to be created to close
            # the surface of the model; its parameters are derived from those of
            # the outer box and adjusted to make both boxes fit together.

            w = inner_dims["x"]
            d = inner_dims["y"]
            h = inner_dims["z"]
            x = (inner_corners["-x"] - inner_corners["x"]) * .5 + center[0]
            y = (inner_corners["-y"] - inner_corners["y"]) * .5 + center[1]
            z = (inner_corners["-z"] - inner_corners["z"]) * .5 + center[2]
            segs = {"width": segs_w, "depth": segs_d, "height": segs_h}
            inner_tex_units = {} if tex_units else None

            if tex_units:
                for side_name in self._side_ids:
                    if "inner_" + side_name in tex_units:
                        inner_tex_units[side_name] = tex_units["inner_" + side_name]

            inner_tex_offset = {} if tex_offset else None

            if tex_offset:
                for side_name in self._side_ids:
                    if "inner_" + side_name in tex_offset:
                        inner_tex_offset[side_name] = tex_offset["inner_" + side_name]

            inner_tex_rot = {} if tex_rotation else None

            if tex_rotation:
                for side_name in self._side_ids:
                    if "inner_" + side_name in tex_rotation:
                        inner_tex_rot[side_name] = tex_rotation["inner_" + side_name]

            inner_tex_scale = {} if tex_scale else None

            if tex_scale:
                for side_name in self._side_ids:
                    if "inner_" + side_name in tex_scale:
                        inner_tex_scale[side_name] = tex_scale["inner_" + side_name]

            model_maker = BoxMaker(
                (x, y, z),
                w,
                d,
                h,
                segs,
                self._open_sides,
                inverted=not inverted,
                has_uvs=has_uvs,
                tex_units=inner_tex_units,
                tex_offset=inner_tex_offset,
                tex_rotation=inner_tex_rot,
                tex_scale=inner_tex_scale
            )
            node = model_maker.generate()

            # Extend the geometry of the inner box with the data of the outer box

            geom = node.modify_geom(0)
            vertex_data = geom.modify_vertex_data()
            old_vert_count = vertex_data.get_num_rows()
            old_size = old_vert_count * stride
            vertex_data.set_num_rows(old_vert_count + len(values) // stride)
            data_array = vertex_data.modify_array(0)
            memview = memoryview(data_array).cast("B").cast("f")
            memview[old_size:] = values

            if self._vertex_color:
                if has_uvs:
                    vertex_format = GeomVertexFormat.get_v3n3c4t2()
                else:
                    vertex_format = GeomVertexFormat.get_v3n3c4()
                vertex_data.set_format(vertex_format)
                vertex_data = vertex_data.set_color(self._vertex_color)
                geom.set_vertex_data(vertex_data)

            tris_prim = geom.modify_primitive(0)
            old_row_count = tris_prim.get_num_vertices()
            new_row_count = old_row_count + len(indices)

            if new_row_count < 2**16:
                # make the array compatible with the default index format of the
                # GeomPrimitive (16-bit) if the number of vertices allows it...
                indices = array.array("H", indices)
            else:
                # ...or force the GeomPrimitive to accept more indices by setting
                # its index format to 32-bit
                tris_prim.set_index_type(Geom.NT_uint32)

            tris_array = tris_prim.modify_vertices()
            tris_array.set_num_rows(new_row_count)
            memview = memoryview(tris_array).cast("B").cast(indices.typecode)
            memview[old_row_count:] = indices
            tris_prim.offset_vertices(old_vert_count, old_row_count, new_row_count)

            inner_ranges = model_maker.vertex_ranges

            for side_name in ("left", "right", "back", "front", "bottom", "top"):

                inner_range = inner_ranges[side_name]

                if inner_range:
                    vert_ranges["inner_{}".format(side_name)] = inner_range

                vert_range = vert_ranges[side_name]

                if vert_range:
                    start, end = vert_range
                    start += old_vert_count
                    end += old_vert_count
                    vert_ranges[side_name] = (start, end)

        else:

            if has_uvs:
                vertex_format = GeomVertexFormat.get_v3n3t2()
            else:
                vertex_format = GeomVertexFormat.get_v3n3()

            vertex_data = GeomVertexData("box_data", vertex_format, Geom.UH_static)
            vertex_data.unclean_set_num_rows(len(values) // stride)
            data_array = vertex_data.modify_array(0)
            memview = memoryview(data_array).cast("B").cast("f")
            memview[:] = values

            if self._vertex_color:
                if has_uvs:
                    vertex_format = GeomVertexFormat.get_v3n3c4t2()
                else:
                    vertex_format = GeomVertexFormat.get_v3n3c4()
                vertex_data.set_format(vertex_format)
                vertex_data = vertex_data.set_color(self._vertex_color)

            tris_prim = GeomTriangles(Geom.UH_static)

            if len(indices) < 2**16:
                indices = array.array("H", indices)
            else:
                tris_prim.set_index_type(Geom.NT_uint32)

            tris_prim = GeomTriangles(Geom.UH_static)
            tris_array = tris_prim.modify_vertices()
            tris_array.unclean_set_num_rows(len(indices))
            memview = memoryview(tris_array).cast("B").cast(indices.typecode)
            memview[:] = indices

            geom = Geom(vertex_data)
            geom.add_primitive(tris_prim)
            node = GeomNode("box_node")
            node.add_geom(geom)

        return node



import os
import pickle
import shutil

from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.policy.replay_policy import WaymoReplayEgoCarPolicy


def test_export_metadrive_scenario(render_export_env=False, render_load_env=False):
    num_scenarios = 3
    env = MetaDriveEnv(
        dict(start_seed=0, use_render=render_export_env, num_scenarios=num_scenarios, agent_policy=IDMPolicy)
    )
    policy = lambda x: [0, 1]
    dir = None
    try:
        scenarios, done_info = env.export_scenarios(policy, scenario_index=[i for i in range(num_scenarios)])
        dir = os.path.join(os.path.dirname(__file__), "../test_component/test_export")
        os.makedirs(dir, exist_ok=True)
        for i, data in scenarios.items():
            with open(os.path.join(dir, "{}.pkl".format(i)), "wb+") as file:
                pickle.dump(data, file)
        env.close()

        env = WaymoEnv(
            dict(
                agent_policy=WaymoReplayEgoCarPolicy,
                data_directory=dir,
                use_render=render_load_env,
                num_scenarios=num_scenarios
            )
        )
        for index in range(num_scenarios):
            env.reset(seed=index)
            done = False
            while not done:
                o, r, tm, tc, i = env.step([0, 0])
                done = tm or tc
    finally:
        env.close()
        if dir is not None:
            shutil.rmtree(dir)


def test_export_waymo_scenario(num_scenarios=3, render_export_env=False, render_load_env=False):
    env = WaymoEnv(
        dict(
            agent_policy=WaymoReplayEgoCarPolicy,
            use_render=render_export_env,
            start_scenario_index=0,
            num_scenarios=num_scenarios
        )
    )
    policy = lambda x: [0, 1]
    dir = None
    try:
        scenarios, done_info = env.export_scenarios(
            policy, scenario_index=[i for i in range(num_scenarios)], verbose=True
        )
        dir = os.path.join(os.path.dirname(__file__), "../test_component/test_export")
        os.makedirs(dir, exist_ok=True)
        for i, data in scenarios.items():
            with open(os.path.join(dir, "{}.pkl".format(i)), "wb+") as file:
                pickle.dump(data, file)
        env.close()

        print("===== Start restoring =====")
        env = WaymoEnv(
            dict(
                agent_policy=WaymoReplayEgoCarPolicy,
                data_directory=dir,
                use_render=render_load_env,
                num_scenarios=num_scenarios
            )
        )
        for index in range(num_scenarios):
            print("Start replaying scenario {}".format(index))
            env.reset(seed=index)
            done = False
            count = 0
            while not done:
                o, r, tm, tc, i = env.step([0, 0])
                count += 1
                done = tm or tc
            print("Finish replaying scenario {} with step {}".format(index, count))
    finally:
        env.close()
        if dir is not None:
            shutil.rmtree(dir)


if __name__ == "__main__":
    # test_export_metadrive_scenario(render_export_env=False, render_load_env=False)
    test_export_waymo_scenario(num_scenarios=3, render_export_env=False, render_load_env=False)



import logging


class CustomFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "[%(levelname)s] %(message)s (%(name)s %(filename)s:%(lineno)d)"
    simple_format = "[%(levelname)s] %(message)s"

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + simple_format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_logger(name, level=logging.INFO, propagate=False):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.propagate = propagate
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    # create formatter and add it to the handlers
    ch.setFormatter(CustomFormatter())
    logger.addHandler(ch)
    return logger



from typing import Sequence, Tuple

from panda3d.core import NodePath

from metadrive.base_class.base_object import BaseObject

LaneIndex = Tuple[str, str, int]


class BaseStaticObject(BaseObject):
    MASS = 1
    HEIGHT = None

    def __init__(self, position: Sequence[float], heading_theta: float = 0., lane=None, random_seed=None, name=None):
        """
        :param lane: the lane to spawn object
        :param position: cartesian position of object in the surface
        :param heading_theta: the angle from positive direction of horizontal axis
        """
        super(BaseStaticObject, self).__init__(random_seed=random_seed, name=name)
        self.set_position(position, self.HEIGHT / 2 if hasattr(self, "HEIGHT") else 0)
        self.set_heading_theta(heading_theta)
        self.lane = lane
        self.lane_index = lane.index if lane is not None else None

    def show_coordinates(self):
        if self.coordinates_debug_np is not None:
            return
        height = self.HEIGHT
        self.coordinates_debug_np = NodePath("debug coordinate")
        x = self.engine.draw_line_3d([0, 0, height], [2, 0, height], [1, 0, 0, 1], 1)
        y = self.engine.draw_line_3d([0, 0, height], [0, 1, height], [1, 0, 0, 1], 1)
        z = self.engine.draw_line_3d([0, 0, height], [0, 0, height + 0.5], [0, 0, 1, 1], 2)
        x.reparentTo(self.coordinates_debug_np)
        y.reparentTo(self.coordinates_debug_np)
        z.reparentTo(self.coordinates_debug_np)
        self.coordinates_debug_np.reparentTo(self.origin)

    def reset(self, position, heading_theta, lane=None, random_seed=None, name=None, *args, **kwargs):
        self.seed(random_seed)
        self.rename(name)
        self.set_position(position, self.HEIGHT / 2 if hasattr(self, "HEIGHT") else 0)
        self.set_heading_theta(heading_theta)
        self.lane = lane
        self.lane_index = lane.index if lane is not None else None



from collections import deque

import gymnasium as gym
import math
import numpy as np

from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.constants import Decoration, DEFAULT_AGENT
from metadrive.obs.top_down_obs import TopDownObservation
from metadrive.obs.top_down_obs_impl import WorldSurface, COLOR_BLACK, VehicleGraphics, LaneGraphics, \
    ObservationWindowMultiChannel
from metadrive.utils import import_pygame, clip

pygame = import_pygame()
COLOR_WHITE = pygame.Color("white")


class TopDownMultiChannel(TopDownObservation):
    """
    Most of the source code is from Highway-Env, we only optimize and integrate it in MetaDrive
    See more information on its Github page: https://github.com/eleurent/highway-env
    """
    RESOLUTION = (100, 100)  # pix x pix
    MAP_RESOLUTION = (2000, 2000)  # pix x pix
    # MAX_RANGE = (50, 50)  # maximum detection distance = 50 M

    # CHANNEL_NAMES = ["road_network", "traffic_flow", "target_vehicle", "navigation", "past_pos"]
    CHANNEL_NAMES = ["road_network", "traffic_flow", "navigation", "past_pos"]

    def __init__(
        self,
        vehicle_config,
        onscreen,
        clip_rgb: bool,
        frame_stack: int = 5,
        post_stack: int = 5,
        frame_skip: int = 5,
        resolution=None,
        max_distance=50
    ):
        super(TopDownMultiChannel, self).__init__(
            vehicle_config, clip_rgb, onscreen=onscreen, resolution=resolution, max_distance=max_distance
        )
        self.num_stacks = 2 + frame_stack
        self.stack_traffic_flow = deque([], maxlen=(frame_stack - 1) * frame_skip + 1)
        self.stack_past_pos = deque(
            [], maxlen=(post_stack - 1) * frame_skip + 1
        )  # In the coordination of target vehicle
        self.frame_stack = frame_stack
        self.frame_skip = frame_skip
        self._should_fill_stack = True
        self.max_distance = max_distance
        self.scaling = self.resolution[0] / max_distance
        assert self.scaling == self.resolution[1] / self.max_distance

    def init_obs_window(self):
        names = self.CHANNEL_NAMES.copy()
        names.remove("past_pos")
        self.obs_window = ObservationWindowMultiChannel(names, (self.max_distance, self.max_distance), self.resolution)

    def init_canvas(self):
        self.canvas_background = WorldSurface(self.MAP_RESOLUTION, 0, pygame.Surface(self.MAP_RESOLUTION))
        self.canvas_navigation = WorldSurface(self.MAP_RESOLUTION, 0, pygame.Surface(self.MAP_RESOLUTION))
        self.canvas_road_network = WorldSurface(self.MAP_RESOLUTION, 0, pygame.Surface(self.MAP_RESOLUTION))
        self.canvas_runtime = WorldSurface(self.MAP_RESOLUTION, 0, pygame.Surface(self.MAP_RESOLUTION))
        self.canvas_ego = WorldSurface(self.MAP_RESOLUTION, 0, pygame.Surface(self.MAP_RESOLUTION))
        self.canvas_past_pos = pygame.Surface(self.resolution)  # A local view

    def reset(self, env, vehicle=None):
        # self.engine = env.engine
        self.road_network = env.current_map.road_network
        self.target_vehicle = vehicle
        self._should_draw_map = True
        self._should_fill_stack = True

    def draw_map(self) -> pygame.Surface:
        """
        :return: a big map surface, clip  and rotate to use a piece of it
        """
        # Setup the maximize size of the canvas
        # scaling and center can be easily found by bounding box
        b_box = self.road_network.get_bounding_box()
        self.canvas_navigation.fill(COLOR_BLACK)
        self.canvas_ego.fill(COLOR_BLACK)
        self.canvas_road_network.fill(COLOR_BLACK)
        self.canvas_runtime.fill(COLOR_BLACK)
        self.canvas_background.fill(COLOR_BLACK)
        self.canvas_background.set_colorkey(self.canvas_background.BLACK)
        x_len = b_box[1] - b_box[0]
        y_len = b_box[3] - b_box[2]
        max_len = max(x_len, y_len) + 20  # Add more 20 meters
        scaling = self.MAP_RESOLUTION[1] / max_len - 0.1
        assert scaling > 0

        # real-world distance * scaling = pixel in canvas
        self.canvas_background.scaling = scaling
        self.canvas_runtime.scaling = scaling
        self.canvas_navigation.scaling = scaling
        self.canvas_ego.scaling = scaling
        self.canvas_road_network.scaling = scaling

        centering_pos = ((b_box[0] + b_box[1]) / 2, (b_box[2] + b_box[3]) / 2)
        self.canvas_runtime.move_display_window_to(centering_pos)
        self.canvas_navigation.move_display_window_to(centering_pos)
        self.canvas_ego.move_display_window_to(centering_pos)
        self.canvas_background.move_display_window_to(centering_pos)
        self.canvas_road_network.move_display_window_to(centering_pos)

        # self.draw_navigation(self.canvas_navigation)
        self.draw_navigation(self.canvas_background, (64, 64, 64))

        for _from in self.road_network.graph.keys():
            decoration = True if _from == Decoration.start else False
            for _to in self.road_network.graph[_from].keys():
                for l in self.road_network.graph[_from][_to]:
                    two_side = True if l is self.road_network.graph[_from][_to][-1] or decoration else False
                    LaneGraphics.LANE_LINE_WIDTH = 0.5
                    LaneGraphics.display(l, self.canvas_background, two_side)
        self.canvas_road_network.blit(self.canvas_background, (0, 0))
        self.obs_window.reset(self.canvas_runtime)
        self._should_draw_map = False

    def _refresh(self, canvas, pos, clip_size):
        canvas.set_clip((pos[0] - clip_size[0] / 2, pos[1] - clip_size[1] / 2, clip_size[0], clip_size[1]))
        canvas.fill(COLOR_BLACK)

    def draw_scene(self):
        # Set the active area that can be modify to accelerate
        assert len(self.engine.agents) == 1, "Don't support multi-agent top-down observation yet!"
        vehicle = self.engine.agents[DEFAULT_AGENT]
        pos = self.canvas_runtime.pos2pix(*vehicle.position)

        clip_size = (int(self.obs_window.get_size()[0] * 1.1), int(self.obs_window.get_size()[0] * 1.1))

        # self._refresh(self.canvas_ego, pos, clip_size)
        self._refresh(self.canvas_runtime, pos, clip_size)
        self.canvas_past_pos.fill(COLOR_BLACK)
        # self._draw_ego_vehicle()

        # Draw vehicles
        # TODO PZH: I hate computing these in pygame-related code!!!
        ego_heading = vehicle.heading_theta
        ego_heading = ego_heading if abs(ego_heading) > 2 * np.pi / 180 else 0

        for v in self.engine.traffic_manager.vehicles:
            if v is vehicle:
                continue
            h = v.heading_theta
            h = h if abs(h) > 2 * np.pi / 180 else 0
            VehicleGraphics.display(vehicle=v, surface=self.canvas_runtime, heading=h, color=VehicleGraphics.BLUE)

        raw_pos = vehicle.position
        self.stack_past_pos.append(raw_pos)
        for p_index in self._get_stack_indices(len(self.stack_past_pos)):
            p_old = self.stack_past_pos[p_index]
            diff = p_old - raw_pos
            diff = (diff[0] * self.scaling, diff[1] * self.scaling)
            # p = (p_old[0] - pos[0], p_old[1] - pos[1])
            diff = (diff[1], diff[0])
            p = pygame.math.Vector2(tuple(diff))
            # p = pygame.math.Vector2(p)
            p = p.rotate(np.rad2deg(ego_heading) + 90)
            p = (p[1], p[0])
            p = (
                clip(p[0] + self.resolution[0] / 2, -self.resolution[0],
                     self.resolution[0]), clip(p[1] + self.resolution[1] / 2, -self.resolution[1], self.resolution[1])
            )
            # p = self.canvas_background.pos2pix(p[0], p[1])
            self.canvas_past_pos.fill((255, 255, 255), (p, (1, 1)))
            # pygame.draw.circle(self.canvas_past_pos, (255, 255, 255), p, radius=1)

        ret = self.obs_window.render(
            canvas_dict=dict(
                road_network=self.canvas_road_network,
                traffic_flow=self.canvas_runtime,
                target_vehicle=self.canvas_ego,
                # navigation=self.canvas_navigation,
            ),
            position=pos,
            heading=vehicle.heading_theta
        )
        ret["past_pos"] = self.canvas_past_pos
        return ret

    def _draw_ego_vehicle(self):
        vehicle = self.engine.agents[DEFAULT_AGENT]
        w = vehicle.WIDTH * self.scaling
        h = vehicle.LENGTH * self.scaling
        position = (self.resolution[0] / 2, self.resolution[1] / 2)
        angle = 90
        box = [pygame.math.Vector2(p) for p in [(-h / 2, -w / 2), (-h / 2, w / 2), (h / 2, w / 2), (h / 2, -w / 2)]]
        box_rotate = [p.rotate(angle) + position for p in box]
        pygame.draw.polygon(self.canvas_past_pos, color=(128, 128, 128), points=box_rotate)

    def get_observation_window(self):
        ret = self.obs_window.get_observation_window()
        ret["past_pos"] = self.canvas_past_pos
        return ret

    def _transform(self, img):
        # img = np.mean(img, axis=-1)
        # Use Atari-like processing

        # img = img[..., 0]
        # img = np.dot(img[..., :], [0.299, 0.587, 0.114])
        img = img[..., 0] * 0.299 + img[..., 1] * 0.587 + img[..., 2] * 0.114

        if self.rgb_clip:
            img = img.astype(np.float32) / 255
        else:
            img = img.astype(np.uint8)
        return img

    def observe(self, vehicle: BaseVehicle):
        self.render()
        surface_dict = self.get_observation_window()
        surface_dict["road_network"] = pygame.transform.smoothscale(surface_dict["road_network"], self.resolution)
        img_dict = {k: pygame.surfarray.array3d(surface) for k, surface in surface_dict.items()}

        # Gray scale
        img_dict = {k: self._transform(img) for k, img in img_dict.items()}

        if self._should_fill_stack:
            self.stack_past_pos.clear()
            self.stack_traffic_flow.clear()
            for _ in range(self.stack_traffic_flow.maxlen):
                self.stack_traffic_flow.append(img_dict["traffic_flow"])
            self._should_fill_stack = False
        self.stack_traffic_flow.append(img_dict["traffic_flow"])

        img = [
            img_dict["road_network"] * 2,
            # img_navigation,
            # img_dict["navigation"],
            # img_dict["target_vehicle"],
            img_dict["past_pos"],
        ]  # + list(self.stack_traffic_flow)

        # Stacked traffic flow
        # stacked = np.zeros_like(img_navigation)
        indices = self._get_stack_indices(len(self.stack_traffic_flow))
        # for i in reversed(indices):
        #     stacked = self.stack_traffic_flow[i] + stacked / 2
        # if self.rgb_clip:
        #     stacked = np.clip(stacked, 0.0, 1.0)
        # else:
        #     stacked = np.clip(stacked, 0, 255)
        for i in indices:
            img.append(self.stack_traffic_flow[i])

        # Stack
        img = np.stack(img, axis=2)
        if self.rgb_clip:
            img = np.clip(img, 0, 1.0)
        else:
            img = np.clip(img, 0, 255)
        return np.transpose(img, (1, 0, 2))

    def draw_navigation(self, canvas, color=(128, 128, 128)):
        checkpoints = self.target_vehicle.navigation.checkpoints
        for i, c in enumerate(checkpoints[:-1]):
            lanes = self.road_network.graph[c][checkpoints[i + 1]]
            for lane in lanes:
                LaneGraphics.draw_drivable_area(lane, canvas, color=color)

    def _get_stack_indices(self, length, frame_skip=None):
        frame_skip = frame_skip or self.frame_skip
        num = int(math.ceil(length / frame_skip))
        indices = []
        for i in range(num):
            indices.append(length - 1 - i * frame_skip)
        return indices

    @property
    def observation_space(self):
        shape = self.obs_shape + (self.num_stacks, )
        if self.rgb_clip:
            return gym.spaces.Box(-0.0, 1.0, shape=shape, dtype=np.float32)
        else:
            return gym.spaces.Box(0, 255, shape=shape, dtype=np.uint8)



import copy
from typing import Union, Any

import numpy as np

from metadrive.utils.utils import merge_dicts


def merge_config_with_unknown_keys(old_dict, new_dict):
    return merge_config(old_dict, new_dict, new_keys_allowed=True)


def merge_config(old_dict, new_dict, new_keys_allowed=False):
    from metadrive.utils import Config
    if isinstance(old_dict, Config):
        old_dict = old_dict.get_dict()
    if isinstance(new_dict, Config):
        new_dict = new_dict.get_dict()
    merged = merge_dicts(old_dict, new_dict, allow_new_keys=new_keys_allowed)
    return Config(merged)


def _check_keys(new_config: Union[dict, "Config"], old_config: Union[dict, "Config"], prefix=""):
    if isinstance(new_config, Config):
        new_config = new_config.get_dict()
    if isinstance(old_config, Config):
        old_config = new_config.get_dict()
    assert isinstance(new_config, dict)
    assert isinstance(old_config, dict)
    own_keys = set(old_config)
    new_keys = set(new_config)
    if own_keys >= new_keys:
        return True
    else:
        raise KeyError(
            "Unexpected keys: {} in new dict{} when update config. Existing keys: {}.".format(
                new_keys - own_keys, "'s '{}'".format(prefix) if prefix else "", own_keys
            )
        )


def _recursive_check_keys(new_config, old_config, prefix=""):
    _check_keys(new_config, old_config, prefix)
    for k, v in new_config.items():
        new_prefix = prefix + "/" + k if prefix else k
        # if isinstance(v, dict):
        #     _recursive_check_keys(new_config[k], old_config[k], new_prefix)
        if isinstance(v, list):
            for new, old in zip(v, old_config[k]):
                _recursive_check_keys(new, old, new_prefix)


def config_to_dict(config: Union[Any, dict, "Config"], serializable=False) -> dict:
    # Return the flatten and json-able dict
    if not isinstance(config, (dict, Config)):
        return config
    ret = dict()
    for k, v in config.items():
        if isinstance(v, Config):
            v = v.get_dict()
        elif isinstance(v, dict):
            v = {sub_k: config_to_dict(sub_v, serializable) for sub_k, sub_v in v.items()}
        elif serializable and isinstance(v, np.ndarray):
            v = v.tolist()
        ret[k] = v
    return ret


class Config:
    """
    This class aims to check whether user config exists if default config system,
    Mostly, Config is sampled from parameter space in MetaDrive

    Besides, the value type will also be checked, but sometimes the value type is not unique (maybe Union[str, int]).
    For these <key, value> items, use Config["your key"] = None to init your PgConfig, then it will not implement
    type check at the first time. key "config" in map.py and key "force_fps" in world.py are good examples.
    """
    def __init__(self, config: Union["Config", dict], unchangeable=False):
        self._unchangeable = False
        if isinstance(config, Config):
            config = config.get_dict()
        self._config = self._internal_dict_to_config(copy.deepcopy(config))
        self._types = dict()
        for k, v in self._config.items():
            self._set_item(k, v, allow_overwrite=True)
        self._unchangeable = unchangeable

    def clear(self):
        self._config.clear()

    def register_type(self, key, *types):
        """
        Register special types for item in config. This is used for mixed type declaration.
        Note that is the type is declared as None, then we will not check type for this item.
        """
        assert key in self._config
        self._types[key] = set(types)

    def get_dict(self):
        return config_to_dict(self._config, serializable=False)

    def get_serializable_dict(self):
        return config_to_dict(self._config, serializable=True)

    def update(self, new_dict: Union[dict, "Config"], allow_add_new_key=True, stop_recursive_update=None):
        """
        Update this dict with extra configs
        :param new_dict: extra configs
        :param allow_add_new_key: whether allowing to add new keys to existing configs or not
        :param stop_recursive_update: Deep update and recursive-check will NOT be applied to keys in stop_recursive_update
        """
        new_dict = new_dict or dict()
        if len(new_dict) == 0:
            return self
        stop_recursive_update = stop_recursive_update or []
        new_dict = copy.deepcopy(new_dict)
        if not allow_add_new_key:
            old_keys = set(self._config)
            new_keys = set(new_dict)
            diff = new_keys.difference(old_keys)
            if len(diff) > 0:
                raise KeyError(
                    "'{}' does not exist in existing config. "
                    "Please use config.update(...) to update the config. Existing keys: {}.".format(
                        diff, self._config.keys()
                    )
                )
        for k, v in new_dict.items():
            if k not in self:
                if isinstance(v, dict):
                    v = Config(v)
                self._config[k] = v  # Placeholder
            success = False
            if isinstance(self._config[k], (dict, Config)):
                if k not in stop_recursive_update:
                    success = self._update_dict_item(k, v, allow_add_new_key)
                else:
                    self._set_item(k, v, allow_add_new_key)
                    success = True
            if not success:
                self._update_single_item(k, v, allow_add_new_key)
            if k in self._config and not hasattr(self, k):
                self.__setattr__(k, self._config[k])
        return self

    def _update_dict_item(self, k, v, allow_overwrite):
        if not isinstance(v, (dict, Config)):
            if allow_overwrite:
                return False
            else:
                raise TypeError(
                    "Type error! The item {} has original type {} and updating type {}.".format(
                        k, type(self[k]), type(v)
                    )
                )
        if not isinstance(self[k], Config):
            self._set_item(k, Config(self[k]), allow_overwrite)
        self[k].update(v, allow_add_new_key=allow_overwrite)
        return True

    def _update_single_item(self, k, v, allow_overwrite):
        assert not isinstance(v, (dict, Config)), (k, type(v), allow_overwrite)
        self._set_item(k, v, allow_overwrite)

    def items(self):
        return self._config.items()

    def values(self):
        return self._config.values()

    def keys(self):
        return self._config.keys()

    def pop(self, key):
        assert hasattr(self, key)
        self._config.pop(key)
        self.__delattr__(key)

    @classmethod
    def _internal_dict_to_config(cls, d: dict) -> dict:
        ret = dict()
        d = d or dict()
        for k, v in d.items():
            if isinstance(v, dict):
                v = cls(v)
            ret[k] = v
        return ret

    def _check_and_raise_key_error(self, key):
        if key not in self._config:
            raise KeyError(
                "'{}' does not exist in existing config. "
                "Please use config.update(...) to update the config. Existing keys: {}.".format(
                    key, self._config.keys()
                )
            )

    def copy(self, unchangeable=None):
        """If unchangeable is None, then just following the original config's setting."""
        if unchangeable is None:
            unchangeable = self._unchangeable
        return Config(self, unchangeable)

    def __getitem__(self, item):
        self._check_and_raise_key_error(item)
        ret = self._config[item]
        return ret

    def _set_item(self, key, value, allow_overwrite):
        """A helper function to replace __setattr__ and __setitem__!"""
        self._check_and_raise_key_error(key)
        if isinstance(value, np.ndarray) and len(value) == 1:
            # handle 1-d box shape sample
            value = value[0]
            if isinstance(value, (np.float32, np.float64, float)):
                value = float(value)
            if isinstance(value, (np.int32, np.int64, np.uint)):
                value = int(value)
        if self._unchangeable:
            raise ValueError("This config is not changeable!")
        if (not allow_overwrite) and (self._config[key] is not None and value is not None):
            type_correct = isinstance(value, type(self._config[key]))
            if isinstance(self._config[key], Config):
                type_correct = type_correct or isinstance(value, dict)
            if isinstance(self._config[key], float):
                # int can be transformed to float
                type_correct = type_correct or isinstance(value, int)
            if isinstance(self._config[key], int):
                # float can be transformed to int
                type_correct = type_correct or isinstance(value, float)
            if key in self._types:
                if None in self._types[key]:
                    type_correct = True
                type_correct = type_correct or (type(value) in self._types[key])
            assert type_correct, "TypeError: {}:{}".format(key, value)
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        if self._unchangeable:
            raise ValueError("This config is not changeable!")
        self._config[key] = value
        super(Config, self).__setattr__(key, value)

    def __setattr__(self, key, value):
        if key not in ["_config", "_types", "_unchangeable"]:
            self.__setitem__(key, value)
        else:
            super(Config, self).__setattr__(key, value)

    def __contains__(self, item):
        return item in self._config

    def __repr__(self):
        return str(self._config)

    def __len__(self):
        return len(self._config)

    def __iter__(self):
        return iter(self.keys())

    def check_keys(self, new_dict):
        if isinstance(new_dict, (dict, Config)):
            for k, v in new_dict.items():
                if k not in self:
                    return False, k
                else:
                    return self.check_keys(v)
        else:
            return True, None

    def remove_keys(self, keys):
        for k in keys:
            if k in self:
                self.pop(k)

    def is_identical(self, new_dict: Union[dict, "Config"]) -> bool:
        assert isinstance(new_dict, (dict, Config))
        return _is_identical("", self, "", new_dict)

    def get(self, key, *args):
        return self._config.get(key, *args)

    def force_update(self, new_config):
        self._unchangeable = False
        self.update(new_config)
        self._unchangeable = True

    def force_set(self, key, value):
        self._unchangeable = False
        self[key] = value
        self._unchangeable = True


def _is_identical(k1, v1, k2, v2):
    if k1 != k2:
        return False
    if isinstance(v1, (dict, Config)) or isinstance(v2, (dict, Config)):
        if (not isinstance(v2, (dict, Config))) or (not isinstance(v1, (dict, Config))):
            return False
        if set(v1.keys()) != set(v2.keys()):
            return False
        for k in v1.keys():
            if not _is_identical(k, v1[k], k, v2[k]):
                return False
    else:
        if v1 != v2:
            return False
    return True


def filter_none(config):
    to_remove = []
    for k, v in config.items():
        if v is None:
            to_remove.append(k)
    for k in to_remove:
        config.pop(k)
    return config



# in order to use pbr in opengles pipe on clusters, we temporally inherit from simple pbr
import panda3d.core as p3d

from panda3d.core import Shader, ConfigVariableString
from simplepbr import Pipeline, _add_shader_defines, _load_shader_str

from metadrive.engine.asset_loader import AssetLoader

# def _load_shader_str(shaderpath, defines=None):
#     shader_dir = AssetLoader.file_path("shaders", "pbr_shaders", shaderpath)
#
#     with open(shader_dir) as shaderfile:
#         shaderstr = shaderfile.read()
#
#     if defines is not None:
#         shaderstr = _add_shader_defines(shaderstr, defines)
#
#     return shaderstr


class OurPipeline(Pipeline):
    # raise DeprecationWarning("This feature is deprecated now")

    def __init__(
        self,
        render_node=None,
        window=None,
        camera_node=None,
        taskmgr=None,
        msaa_samples=4,
        max_lights=8,
        use_normal_maps=False,
        use_emission_maps=True,
        exposure=1.0,
        enable_shadows=False,
        enable_fog=False,
        use_occlusion_maps=False
    ):
        super(OurPipeline, self).__init__(
            render_node=render_node,
            window=window,
            camera_node=camera_node,
            taskmgr=taskmgr,
            msaa_samples=msaa_samples,
            max_lights=max_lights,
            use_normal_maps=use_normal_maps,
            use_emission_maps=use_emission_maps,
            exposure=exposure,
            enable_shadows=enable_shadows,
            enable_fog=enable_fog,
            use_occlusion_maps=use_occlusion_maps
        )

    def _setup_tonemapping(self):
        # return
        if self._shader_ready:
            # Destroy previous buffers so we can re-create
            self.manager.cleanup()

            # Fix shadow buffers after FilterManager.cleanup()
            for caster in self.get_all_casters():
                sbuff_size = caster.get_shadow_buffer_size()
                caster.set_shadow_buffer_size((0, 0))
                caster.set_shadow_buffer_size(sbuff_size)

        fbprops = p3d.FrameBufferProperties()
        fbprops.float_color = True
        fbprops.set_rgba_bits(16, 16, 16, 16)
        fbprops.set_depth_bits(24)
        fbprops.set_multisamples(self.msaa_samples)
        scene_tex = p3d.Texture()
        scene_tex.set_format(p3d.Texture.F_rgba16)
        scene_tex.set_component_type(p3d.Texture.T_float)
        self.tonemap_quad = self.manager.render_scene_into(colortex=scene_tex, fbprops=fbprops)

        defines = {}
        if self.use_330:
            defines['USE_330'] = ''

        post_vert_str = _load_shader_str('post.vert', defines)
        post_frag_str = _load_shader_str('tonemap.frag', defines)
        tonemap_shader = p3d.Shader.make(
            p3d.Shader.SL_GLSL,
            vertex=post_vert_str,
            fragment=post_frag_str,
        )
        self.tonemap_quad.set_shader(tonemap_shader)
        self.tonemap_quad.set_shader_input('tex', scene_tex)
        self.tonemap_quad.set_shader_input('exposure', self.exposure)

    def _recompile_pbr(self):
        gles = ConfigVariableString("load-display").getValue()
        if gles == "pandagles2":
            pbr_defines = {
                'MAX_LIGHTS': self.max_lights,
            }
            if self.use_normal_maps:
                pbr_defines['USE_NORMAL_MAP'] = ''
            if self.use_emission_maps:
                pbr_defines['USE_EMISSION_MAP'] = ''
            if self.enable_shadows:
                pbr_defines['ENABLE_SHADOWS'] = ''
            if self.enable_fog:
                pbr_defines['ENABLE_FOG'] = ''
            if self.use_occlusion_maps:
                pbr_defines['USE_OCCLUSION_MAP'] = ''

            pbr_vert_str = _load_shader_str('simplepbr_gles.vert', pbr_defines)
            pbr_frag_str = _load_shader_str('simplepbr_gles.frag', pbr_defines)
            pbrshader = Shader.make(
                Shader.SL_GLSL,
                vertex=pbr_vert_str,
                fragment=pbr_frag_str,
            )
            self.render_node.set_shader(pbrshader)
        else:
            super(OurPipeline, self)._recompile_pbr()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division

import os
import sys
import subprocess

from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error

from panda3d.core import Vec4, Vec3, Vec2, RenderState, TransformState
from panda3d.core import TexturePool, SceneGraphAnalyzer
from direct.interval.IntervalGlobal import Sequence

from metadrive.render_pipeline.rpcore.gui.sprite import Sprite
from metadrive.render_pipeline.rpcore.gui.buffer_viewer import BufferViewer
from metadrive.render_pipeline.rpcore.gui.pipe_viewer import PipeViewer
from metadrive.render_pipeline.rpcore.gui.render_mode_selector import RenderModeSelector

from metadrive.render_pipeline.rpcore.gui.text_node import TextNode
from metadrive.render_pipeline.rpcore.gui.error_message_display import ErrorMessageDisplay
from metadrive.render_pipeline.rpcore.gui.exposure_widget import ExposureWidget
from metadrive.render_pipeline.rpcore.gui.fps_chart import FPSChart
from metadrive.render_pipeline.rpcore.gui.pixel_inspector import PixelInspector

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.rpobject import RPObject

from metadrive.render_pipeline.rpcore.native import NATIVE_CXX_LOADED
from metadrive.render_pipeline.rpcore.render_target import RenderTarget
from metadrive.render_pipeline.rpcore.image import Image


class Debugger(RPObject):
    """ This class manages the onscreen control, and displays statistics. """
    def __init__(self, pipeline):
        RPObject.__init__(self)
        self.debug("Creating debugger")
        self.pipeline = pipeline
        self.analyzer = SceneGraphAnalyzer()

        self.fullscreen_node = Globals.base.pixel2d.attach_new_node("rp_debugger")
        self.create_components()
        self.init_keybindings()

        if self.advanced_info:
            Globals.base.doMethodLater(
                0.5, lambda task: self.collect_scene_data(), "RPDebugger_collectSceneData_initial"
            )

        Globals.base.doMethodLater(0.1, self.update_stats, "RPDebugger_updateStats")

    @property
    def advanced_info(self):
        return self.pipeline.settings["pipeline.advanced_debugging_info"]

    def create_components(self):
        """ Creates the gui components """

        self.debugger_width = 460
        self.debugger_visible = False
        self.debugger_interval = None

        self.create_stats()
        self.create_hints()

        self.pipeline_logo = Sprite(
            image="/$$rp/data/gui/pipeline_logo_text.png", x=30, y=30, parent=self.fullscreen_node
        )

        if self.advanced_info:
            self.exposure_node = self.fullscreen_node.attach_new_node("ExposureWidget")
            self.exposure_widget = ExposureWidget(self.pipeline, self.exposure_node)

        self.fps_node = self.fullscreen_node.attach_new_node("FPSChart")
        self.fps_node.set_pos(Vec3(21, 1, -108 - 40))
        self.fps_widget = FPSChart(self.pipeline, self.fps_node)

        self.pixel_widget = PixelInspector(self.pipeline)
        self.buffer_viewer = BufferViewer(self.pipeline, self.fullscreen_node)
        self.pipe_viewer = PipeViewer(self.pipeline, self.fullscreen_node)
        self.rm_selector = RenderModeSelector(self.pipeline, self.fullscreen_node)
        self.error_msg_handler = ErrorMessageDisplay()

        self.handle_window_resize()

    def update(self):
        """ Updates the gui """
        self.error_msg_handler.update()
        self.pixel_widget.update()

    def collect_scene_data(self, task=None):
        """ Analyzes the scene graph to provide useful information """
        self.analyzer.clear()
        for geom_node in Globals.base.render.find_all_matches("**/+GeomNode"):
            self.analyzer.add_node(geom_node.node())
        if task:
            return task.again

    def create_stats(self):
        """ Creates the stats overlay """
        self.overlay_node = Globals.base.aspect2d.attach_new_node("Overlay")
        self.debug_lines = []

        num_lines = 6 if self.advanced_info else 1
        for i in range(num_lines):
            self.debug_lines.append(
                TextNode(pos=Vec2(0, -i * 0.046), parent=self.overlay_node, align="right", color=Vec3(0.7, 1, 1))
            )
        self.debug_lines[0].color = Vec4(1, 1, 0, 1)

    def create_hints(self):
        """ Creates the hints like keybindings and when reloading shaders """
        self.hint_reloading = Sprite(image="/$$rp/data/gui/shader_reload_hint.png", parent=self.fullscreen_node)
        self.set_reload_hint_visible(False)

        self.python_warning = None
        if not NATIVE_CXX_LOADED:
            # Warning when using the python version
            self.python_warning = Sprite(image="/$$rp/data/gui/python_warning.png", parent=self.fullscreen_node)
            Sequence(
                self.python_warning.color_scale_interval(0.7, Vec4(0.3, 1, 1, 0.7), blendType="easeOut"),
                self.python_warning.color_scale_interval(0.7, Vec4(1, 1, 1, 1.0), blendType="easeOut"),
            ).loop()

        # Keybinding hints
        self.keybinding_instructions = Sprite(
            image="/$$rp/data/gui/keybindings.png", parent=self.fullscreen_node, any_filter=False
        )

    def set_reload_hint_visible(self, flag):
        """ Sets whether the shader reload hint is visible """
        if flag:
            self.hint_reloading.show()
        else:
            self.hint_reloading.hide()

    def handle_window_resize(self):
        """ Handles the window resize, repositions the GUI elements to fit on
        screen. """
        # When using small resolutions, scale the GUI so its still useable,
        # otherwise the sub-windows are bigger than the main window
        self.gui_scale = max(0.65, min(1.0, Globals.native_resolution.x / 1920.0))
        self.fullscreen_node.set_scale(self.gui_scale)

        if self.advanced_info:
            self.exposure_node.set_pos(
                Globals.native_resolution.x // self.gui_scale - 200, 1,
                -Globals.native_resolution.y // self.gui_scale + 120
            )
        self.hint_reloading.set_pos(float((Globals.native_resolution.x) // 2) / self.gui_scale - 465 // 2, 220)
        self.keybinding_instructions.set_pos(
            30,
            Globals.native_resolution.y // self.gui_scale - 510.0,
        )
        self.overlay_node.set_pos(Globals.base.get_aspect_ratio() - 0.07, 1, 1.0 - 0.07)
        if self.python_warning:
            self.python_warning.set_pos(
                (Globals.native_resolution.x // self.gui_scale - 1054) // 2,
                (Globals.native_resolution.y // self.gui_scale - 118 - 40)
            )

        for text in self.debug_lines:
            text.set_pixel_size(16 * max(0.8, self.gui_scale))

        self.buffer_viewer.center_on_screen()
        self.pipe_viewer.center_on_screen()
        self.rm_selector.center_on_screen()

    def init_keybindings(self):
        """ Inits the debugger keybindings """
        Globals.base.accept("v", self.buffer_viewer.toggle)
        # Globals.base.accept("c", self.pipe_viewer.toggle)
        Globals.base.accept("z", self.rm_selector.toggle)
        Globals.base.accept("f5", self.toggle_gui_visible)
        Globals.base.accept("f6", self.toggle_keybindings_visible)
        Globals.base.accept("5", self.pipeline.reload_shaders)
        Globals.base.accept("m", self.start_material_editor)

    def start_material_editor(self):
        """ Starts the material editor """
        self.debug("Starting material editor")
        pth = sys.executable
        editor = os.path.dirname(os.path.realpath(__file__))
        editor = os.path.join(editor, "..", "..", "toolkit", "material_editor", "main.py")
        subprocess.Popen([pth, editor], shell=True)

    def toggle_gui_visible(self):
        """ Shows / Hides the gui """

        if not self.fullscreen_node.is_hidden():
            self.fullscreen_node.hide()
            self.overlay_node.hide()
            self.error_msg_handler.hide()
        else:
            self.fullscreen_node.show()
            self.overlay_node.show()
            self.error_msg_handler.show()

    def toggle_keybindings_visible(self):
        """ Shows / Hides the FPS graph """
        if not self.keybinding_instructions.is_hidden():
            self.keybinding_instructions.hide()
        else:
            self.keybinding_instructions.show()

    def update_stats(self, task=None):
        """ Updates the stats overlay """
        clock = Globals.clock
        self.debug_lines[0].text = "{:3.0f} fps  |  {:3.1f} ms  |  {:3.1f} ms max".format(
            clock.get_average_frame_rate(), 1000.0 / max(0.001, clock.get_average_frame_rate()),
            clock.get_max_frame_duration() * 1000.0
        )

        if not self.advanced_info:
            return task.again if task else None

        text = "{:4d} states |  {:4d} transforms "
        text += "|  {:4d} cmds |  {:4d} lights |  {:4d} shadows "
        text += "|  {:5.1f}% atlas usage"
        self.debug_lines[1].text = text.format(
            RenderState.get_num_states(), TransformState.get_num_states(),
            self.pipeline.light_mgr.cmd_queue.num_processed_commands, self.pipeline.light_mgr.num_lights,
            self.pipeline.light_mgr.num_shadow_sources, self.pipeline.light_mgr.shadow_atlas_coverage
        )

        text = "Internal:  {:3.0f} MB VRAM |  {:5d} img |  {:5d} tex |  "
        text += "{:5d} fbos |  {:3d} plugins |  {:2d}  views  ({:2d} active)"
        tex_memory, tex_count = self.buffer_viewer.stage_information

        views, active_views = 0, 0
        for target in RenderTarget.REGISTERED_TARGETS:
            if not target.create_default_region:
                num_regions = target.internal_buffer.get_num_display_regions()
                for i, region in enumerate(target.internal_buffer.get_display_regions()):
                    # Skip overlay display region
                    if i == 0 and num_regions > 1:
                        continue
                    views += 1
                    active_views += 1 if target.active and region.active else 0

        self.debug_lines[2].text = text.format(
            tex_memory / (1024**2), len(Image.REGISTERED_IMAGES), tex_count, RenderTarget.NUM_ALLOCATED_BUFFERS,
            len(self.pipeline.plugin_mgr.enabled_plugins), views, active_views
        )

        text = "Scene:   {:4.0f} MB VRAM |  {:3d} tex |  {:4d} geoms "
        text += "|  {:4d} nodes |  {:7,.0f} vertices"
        scene_tex_size = 0
        for tex in TexturePool.find_all_textures():
            scene_tex_size += tex.estimate_texture_memory()

        self.debug_lines[3].text = text.format(
            scene_tex_size / (1024**2),
            len(TexturePool.find_all_textures()),
            self.analyzer.get_num_geoms(),
            self.analyzer.get_num_nodes(),
            self.analyzer.get_num_vertices(),
        )

        sun_vector = Vec3(0)
        if self.pipeline.plugin_mgr.is_plugin_enabled("scattering"):
            sun_vector = self.pipeline.plugin_mgr.instances["scattering"].sun_vector

        text = "Time:  {} ({:1.3f}) |  Sun  {:0.2f} {:0.2f} {:0.2f}"
        text += " |  X {:3.1f}  Y {:3.1f}  Z {:3.1f}"
        text += " |  {:2d} tasks |  scheduled: {:2d}"
        self.debug_lines[4].text = text.format(
            self.pipeline.daytime_mgr.formatted_time,
            self.pipeline.daytime_mgr.time,
            sun_vector.x,
            sun_vector.y,
            sun_vector.z,
            Globals.base.camera.get_x(Globals.base.render),
            Globals.base.camera.get_y(Globals.base.render),
            Globals.base.camera.get_z(Globals.base.render),
            self.pipeline.task_scheduler.num_tasks,
            self.pipeline.task_scheduler.num_scheduled_tasks,
        )

        text = "Scene shadows:  "
        if "pssm" in self.pipeline.plugin_mgr.enabled_plugins:
            focus = self.pipeline.plugin_mgr.instances["pssm"].scene_shadow_stage.last_focus
            if focus is not None:
                text += "{:3.1f} {:3.1f} {:3.1f} r {:3.1f}".format(focus[0].x, focus[0].y, focus[0].z, focus[1])
            else:
                text += "none"
        else:
            text += "inactive"

        text += "   |  HPR  ({:3.1f}, {:3.1f}, {:3.1f})  |   {:4d} x {:4d} pixels @ {:3.1f} %"
        text += "   |  {:3d} x {:3d} tiles"
        self.debug_lines[5].text = text.format(
            Globals.base.camera.get_h(Globals.base.render),
            Globals.base.camera.get_p(Globals.base.render),
            Globals.base.camera.get_r(Globals.base.render),
            Globals.native_resolution.x,
            Globals.native_resolution.y,
            self.pipeline.settings["pipeline.resolution_scale"] * 100.0,
            self.pipeline.light_mgr.num_tiles.x,
            self.pipeline.light_mgr.num_tiles.y,
        )
        if task:
            return task.again



import logging

import numpy as np
from panda3d.core import TransparencyAttrib, LineSegs, NodePath

from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.constants import RENDER_MODE_ONSCREEN, CamMask
from metadrive.engine.asset_loader import AssetLoader
from metadrive.utils import get_np_random
from metadrive.utils.coordinates_shift import panda_vector


class BaseNavigation:
    """
    Implement all NotImplemented method for customizing a new navigation module.
    This module interact with the map for finding lanes or expected positions
    """
    NUM_WAY_POINT = 2
    CHECK_POINT_INFO_DIM = 5
    NAVI_POINT_DIST = 50
    PRE_NOTIFY_DIST = 40
    MIN_ALPHA = 0.15
    CKPT_UPDATE_RANGE = 5
    FORCE_CALCULATE = False
    LINE_TO_DEST_HEIGHT = 0.6

    def __init__(
        self,
        show_navi_mark: bool = False,
        random_navi_mark_color=False,
        show_dest_mark=False,
        show_line_to_dest=False,
        panda_color=None,
        name=None,
        vehicle_config=None
    ):
        """
        This class define a helper for localizing vehicles and retrieving navigation information.
        It now only support from first block start to the end node, but can be extended easily.
        """
        self.name = name

        # Make sure these variables are filled when making new subclass
        # self.checkpoints = None
        # self.current_ref_lanes = None
        # self.next_ref_lanes = None
        # self.final_lane = None
        # self.current_lane = None

        self.vehicle_config = vehicle_config

        self._target_checkpoints_index = None
        self._navi_info = np.zeros((self.get_navigation_info_dim(), ), dtype=np.float32)  # navi information res

        # Vis TODO make it beautiful!
        self._show_navi_info = (
            self.engine.mode == RENDER_MODE_ONSCREEN and not self.engine.global_config["debug_physics_world"]
        )
        self.origin = NodePath("navigation_sign") if self._show_navi_info else None
        self.navi_mark_color = (0.6, 0.8, 0.5) if not random_navi_mark_color else get_np_random().rand(3)
        if panda_color is not None:
            assert len(panda_color) == 3 and 0 <= panda_color[0] <= 1
            self.navi_mark_color = tuple(panda_color)
        self.navi_arrow_dir = [0, 0]
        self._dest_node_path = None
        self._goal_node_path = None

        self._node_path_list = []

        self._line_to_dest = None
        self._line_to_navi = None
        self._show_line_to_dest = show_line_to_dest
        if self._show_navi_info:
            # nodepath
            self._line_to_dest = self.origin.attachNewNode("line")
            self._goal_node_path = self.origin.attachNewNode("target")
            self._dest_node_path = self.origin.attachNewNode("dest")

            self._node_path_list.append(self._line_to_dest)
            self._node_path_list.append(self._goal_node_path)
            self._node_path_list.append(self._dest_node_path)

            if show_navi_mark:
                navi_point_model = AssetLoader.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
                navi_point_model.reparentTo(self._goal_node_path)
            if show_dest_mark:
                dest_point_model = AssetLoader.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
                dest_point_model.reparentTo(self._dest_node_path)
            if show_line_to_dest:
                line_seg = LineSegs("line_to_dest")
                line_seg.setColor(self.navi_mark_color[0], self.navi_mark_color[1], self.navi_mark_color[2], 1.0)
                line_seg.setThickness(4)
                self._dynamic_line_np = NodePath(line_seg.create(True))

                self._node_path_list.append(self._dynamic_line_np)

                self._dynamic_line_np.reparentTo(self.origin)
                self._line_to_dest = line_seg

            show_line_to_navi_mark = self.vehicle_config["show_line_to_navi_mark"]
            self._show_line_to_navi_mark = show_line_to_navi_mark
            if show_line_to_navi_mark:
                line_seg = LineSegs("line_to_dest")
                line_seg.setColor(self.navi_mark_color[0], self.navi_mark_color[1], self.navi_mark_color[2], 1.0)
                line_seg.setThickness(4)
                self._dynamic_line_np_2 = NodePath(line_seg.create(True))

                self._node_path_list.append(self._dynamic_line_np_2)

                self._dynamic_line_np_2.reparentTo(self.origin)
                self._line_to_navi = line_seg

            self._goal_node_path.setTransparency(TransparencyAttrib.M_alpha)
            self._dest_node_path.setTransparency(TransparencyAttrib.M_alpha)

            self._goal_node_path.setColor(
                self.navi_mark_color[0], self.navi_mark_color[1], self.navi_mark_color[2], 0.7
            )
            self._dest_node_path.setColor(
                self.navi_mark_color[0], self.navi_mark_color[1], self.navi_mark_color[2], 0.7
            )
            self._goal_node_path.hide(CamMask.AllOn)
            self._dest_node_path.hide(CamMask.AllOn)
            self._goal_node_path.show(CamMask.MainCam)
            self._dest_node_path.show(CamMask.MainCam)
        logging.debug("Load Vehicle Module: {}".format(self.__class__.__name__))

    def reset(self, current_lane, vehicle_config=None):
        self._current_lane = current_lane
        if vehicle_config is not None:
            self.vehicle_config = vehicle_config

    @property
    def map(self):
        return self.engine.current_map

    @property
    def current_lane(self):
        return self._current_lane

    def get_checkpoints(self):
        """Return next checkpoint and the next next checkpoint"""
        later_middle = (float(self.get_current_lane_num()) / 2 - 0.5) * self.get_current_lane_width()
        ref_lane1 = self.current_ref_lanes[0]
        checkpoint1 = ref_lane1.position(ref_lane1.length, later_middle)
        ref_lane2 = self.next_ref_lanes[0] if self.next_ref_lanes is not None else self.current_ref_lanes[0]
        checkpoint2 = ref_lane2.position(ref_lane2.length, later_middle)
        return checkpoint1, checkpoint2

    def set_route(self, current_lane_index: str, destination: str):
        """
        Find a shortest path from start road to end road
        :param current_lane_index: start road node
        :param destination: end road node or end lane index
        :return: None
        """
        raise NotImplementedError

    def update_localization(self, ego_vehicle):
        """
        It is called every step. This is the core function of navigation module
        """
        raise NotImplementedError

    def get_navi_info(self):
        return self._navi_info

    def destroy(self):
        if self._show_navi_info:
            try:
                if self._line_to_dest is not None:
                    self._line_to_dest.removeNode()
                if self._line_to_navi is not None:
                    self._line_to_navi.removeNode()
            except AttributeError:
                pass
            self._dest_node_path.removeNode()
            self._goal_node_path.removeNode()

        for np in self._node_path_list:
            np.detachNode()
            np.removeNode()
        # self.next_ref_lanes = None
        # self.current_ref_lanes = None

    def set_force_calculate_lane_index(self, force: bool):
        self.FORCE_CALCULATE = force

    def __del__(self):
        self.destroy()
        logging.debug("{} is destroyed".format(self.__class__.__name__))

    def get_current_lateral_range(self, current_position, engine) -> float:
        raise NotImplementedError

    def get_current_lane_width(self) -> float:
        return self.current_lane.width

    def get_current_lane_num(self) -> float:
        return len(self.current_ref_lanes)

    def _ray_lateral_range(self, engine, start_position, dir, length=50):
        """
        It is used to measure the lateral range of special blocks
        :param start_position: start_point
        :param dir: ray direction
        :param length: length of ray
        :return: lateral range [m]
        """
        end_position = start_position[0] + dir[0] * length, start_position[1] + dir[1] * length
        start_position = panda_vector(start_position, z=0.2)
        end_position = panda_vector(end_position, z=0.2)
        mask = FirstPGBlock.CONTINUOUS_COLLISION_MASK
        res = engine.physics_world.static_world.rayTestClosest(start_position, end_position, mask=mask)
        if not res.hasHit():
            return length
        else:
            return res.getHitFraction() * length

    def _draw_line_to_dest(self, start_position, end_position):
        if not self._show_line_to_dest:
            return
        line_seg = self._line_to_dest
        line_seg.moveTo(panda_vector(start_position, self.LINE_TO_DEST_HEIGHT))
        line_seg.drawTo(panda_vector(end_position, self.LINE_TO_DEST_HEIGHT))
        self._dynamic_line_np.removeNode()
        self._dynamic_line_np = NodePath(line_seg.create(False))

        self._node_path_list.append(self._dynamic_line_np)

        self._dynamic_line_np.hide(CamMask.Shadow | CamMask.RgbCam)
        self._dynamic_line_np.reparentTo(self.origin)

    def _draw_line_to_navi(self, start_position, end_position):
        if not self._show_line_to_navi_mark:
            return
        line_seg = self._line_to_navi
        line_seg.moveTo(panda_vector(start_position, self.LINE_TO_DEST_HEIGHT))
        line_seg.drawTo(panda_vector(end_position, self.LINE_TO_DEST_HEIGHT))
        self._dynamic_line_np_2.removeNode()
        self._dynamic_line_np_2 = NodePath(line_seg.create(False))

        self._node_path_list.append(self._dynamic_line_np_2)

        self._dynamic_line_np_2.hide(CamMask.Shadow | CamMask.RgbCam)
        self._dynamic_line_np_2.reparentTo(self.origin)

    def detach_from_world(self):
        if isinstance(self.origin, NodePath):
            self.origin.detachNode()

    def attach_to_world(self, engine):
        if isinstance(self.origin, NodePath):
            self.origin.reparentTo(engine.render)

    @property
    def engine(self):
        from metadrive.engine.engine_utils import get_engine
        return get_engine()

    def get_state(self):
        return {}

    @classmethod
    def get_navigation_info_dim(cls):
        return cls.NUM_WAY_POINT * cls.CHECK_POINT_INFO_DIM
